#! /usr/bin/env python
"""SearchSource: LocalIndex"""
from __future__ import annotations

import difflib
import typing
import webbrowser
from dataclasses import dataclass
from multiprocessing import Lock
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

import git
import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.load_utils_bib
import colrev.ops.search
import colrev.record
import colrev.ui_cli.cli_colors as colors

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class LocalIndexSearchSource(JsonSchemaMixin):
    """Performs a search in the LocalIndex"""

    # pylint: disable=too-many-instance-attributes
    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "curation_ID"
    search_type = colrev.settings.SearchType.OTHER
    endpoint = "colrev.local_index"
    api_search_supported = True
    ci_supported: bool = True
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "LocalIndex"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/local_index.md"
    )
    __local_index_md_filename = Path("data/search/md_curated.bib")

    essential_md_keys = [
        "title",
        "author",
        "journal",
        "year",
        "booktitle",
        "number",
        "volume",
        "issue",
        "author",
        "doi",
        "colrev.dblp.dblp_key",
        "url",
    ]

    def __init__(
        self,
        *,
        source_operation: colrev.operation.Operation,
        settings: Optional[dict] = None,
    ) -> None:
        if settings:
            # LocalIndex as a search_source
            self.search_source = from_dict(
                data_class=self.settings_class, data=settings
            )

        else:
            # LocalIndex as an md-prep source
            li_md_source_l = [
                s
                for s in source_operation.review_manager.settings.sources
                if s.filename == self.__local_index_md_filename
            ]
            if li_md_source_l:
                self.search_source = li_md_source_l[0]
            else:
                self.search_source = colrev.settings.SearchSource(
                    endpoint=self.endpoint,
                    filename=self.__local_index_md_filename,
                    search_type=colrev.settings.SearchType.OTHER,
                    search_parameters={},
                    comment="",
                )

            self.local_index_lock = Lock()

        self.origin_prefix = self.search_source.get_origin_prefix()

        self.local_index = source_operation.review_manager.get_local_index()
        self.review_manager = source_operation.review_manager

    def __validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""
        source = self.search_source
        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        # if "query" not in source.search_parameters:
        # Note :  for md-sources, there is no query parameter.
        #     raise colrev_exceptions.InvalidQueryException(
        #         f"Source missing query search_parameter ({source.filename})"
        #     )

        if "query" in source.search_parameters:
            pass
            # if "simple_query_string" in source.search_parameters["query"]:
            #     if "query" in source.search_parameters["query"]["simple_query_string"]:
            #         pass
            #     else:
            #         raise colrev_exceptions.InvalidQueryException(
            #             "Source missing query/simple_query_string/query "
            #             f"search_parameter ({source.filename})"
            #         )

            # elif "url" in source.search_parameters["query"]:
            #     pass
            # # else:
            #     raise colrev_exceptions.InvalidQueryException(
            #         f"Source missing query/query search_parameter ({source.filename})"
            #     )

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def __retrieve_from_index(self) -> typing.List[dict]:
        params = self.search_source.search_parameters
        query = params["query"]

        # TODO : generally prefix with SQL fields
        if not any(x in query for x in ["title", "abstract"]):
            query = f'title LIKE "%{query}%"'

        returned_records = self.local_index.search(query=query)

        records_to_import = [r.get_data() for r in returned_records]
        records_to_import = [r for r in records_to_import if r]
        keys_to_drop = [
            "colrev_status",
            "colrev_origin",
            "screening_criteria",
        ]
        for record_dict in records_to_import:
            record_dict = {
                key: value
                for key, value in record_dict.items()
                if key not in keys_to_drop
            }

        return records_to_import

    def __run_md_search_update(
        self,
        *,
        search_operation: colrev.ops.search.Search,
        local_index_feed: colrev.ops.search_feed.GeneralOriginFeed,
    ) -> None:
        records = search_operation.review_manager.dataset.load_records_dict()

        for feed_record_dict_id in list(local_index_feed.feed_records.keys()):
            feed_record_dict = local_index_feed.feed_records[feed_record_dict_id]
            feed_record = colrev.record.Record(data=feed_record_dict)

            try:
                retrieved_record_dict = self.local_index.retrieve(
                    record_dict=feed_record.get_data(), include_file=False
                )

                local_index_feed.set_id(record_dict=retrieved_record_dict)
            except (
                colrev_exceptions.RecordNotInIndexException,
                colrev_exceptions.NotFeedIdentifiableException,
            ):
                continue

            prev_record_dict_version = {}
            if retrieved_record_dict["ID"] in local_index_feed.feed_records:
                prev_record_dict_version = local_index_feed.feed_records[
                    retrieved_record_dict["ID"]
                ]

            local_index_feed.add_record(
                record=colrev.record.Record(data=retrieved_record_dict)
            )

            local_index_feed.update_existing_record(
                records=records,
                record_dict=retrieved_record_dict,
                prev_record_dict_version=prev_record_dict_version,
                source=self.search_source,
                update_time_variant_fields=True,
            )

        local_index_feed.print_post_run_search_infos(records=records)
        local_index_feed.save_feed_file()
        search_operation.review_manager.dataset.save_records_dict(records=records)
        search_operation.review_manager.dataset.add_record_changes()

    def __run_parameter_search(
        self,
        *,
        search_operation: colrev.ops.search.Search,
        local_index_feed: colrev.ops.search_feed.GeneralOriginFeed,
        rerun: bool,
    ) -> None:
        records = search_operation.review_manager.dataset.load_records_dict()

        for retrieved_record_dict in self.__retrieve_from_index():
            try:
                local_index_feed.set_id(record_dict=retrieved_record_dict)
            except colrev_exceptions.NotFeedIdentifiableException:
                continue

            prev_record_dict_version = {}
            if retrieved_record_dict["ID"] in local_index_feed.feed_records:
                prev_record_dict_version = local_index_feed.feed_records[
                    retrieved_record_dict["ID"]
                ]

            added = local_index_feed.add_record(
                record=colrev.record.Record(data=retrieved_record_dict)
            )
            if added:
                self.review_manager.logger.info(
                    " retrieve " + retrieved_record_dict["ID"]
                )

            else:
                local_index_feed.update_existing_record(
                    records=records,
                    record_dict=retrieved_record_dict,
                    prev_record_dict_version=prev_record_dict_version,
                    source=self.search_source,
                    update_time_variant_fields=rerun,
                )

        local_index_feed.print_post_run_search_infos(records=records)
        local_index_feed.save_feed_file()

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of local-index"""

        self.__validate_source()

        local_index_feed = self.search_source.get_feed(
            review_manager=search_operation.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        if self.search_source.is_md_source() or self.search_source.is_quasi_md_source():
            self.__run_md_search_update(
                search_operation=search_operation,
                local_index_feed=local_index_feed,
            )

        else:
            self.__run_parameter_search(
                search_operation=search_operation,
                local_index_feed=local_index_feed,
                rerun=rerun,
            )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for local-index"""

        result = {"confidence": 0.0}
        if "curation_ID" in data:
            result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""

        if params is None:
            operation.add_interactively(endpoint=cls.endpoint)
            return

        filename = operation.get_unique_filename(
            file_path_string=f"local_index_{params}".replace("%", "").replace("'", "")
        )
        add_source = colrev.settings.SearchSource(
            endpoint=cls.endpoint,
            filename=filename,
            search_type=colrev.settings.SearchType.DB,
            search_parameters={"query": params},
            comment="",
        )
        operation.review_manager.settings.sources.append(add_source)

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            for record_id in records:
                records[record_id] = {
                    k: v
                    for k, v in records[record_id].items()
                    if k not in ["colrev_status", "colrev_masterdata_provenance"]
                }

            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for local-index"""

        return record

    def __add_cpid(self, *, record: colrev.record.Record) -> bool:
        # To enable retrieval based on colrev_pdf_id (as part of the global_ids)
        if "file" in record.data and "colrev_pdf_id" not in record.data:
            pdf_path = Path(self.review_manager.path / Path(record.data["file"]))
            if pdf_path.is_file():
                try:
                    record.data.update(
                        colrev_pdf_id=colrev.record.Record.get_colrev_pdf_id(
                            pdf_path=pdf_path,
                        )
                    )
                    return True
                except colrev_exceptions.PDFHashError:
                    pass
        return False

    def __retrieve_record_from_local_index(
        self,
        *,
        record: colrev.record.Record,
        retrieval_similarity: float,
    ) -> colrev.record.Record:
        # add colrev_pdf_id
        added_colrev_pdf_id = self.__add_cpid(record=record)

        retrieved_record_dict = {}
        try:
            retrieved_record_dict = self.local_index.retrieve(
                record_dict=record.get_data(), include_file=False
            )
        except (
            colrev_exceptions.RecordNotInIndexException,
            colrev_exceptions.NotEnoughDataToIdentifyException,
        ):
            try:
                # Search within the table-of-content in local_index
                retrieved_record_dict = self.local_index.retrieve_from_toc(
                    record_dict=record.data,
                    similarity_threshold=retrieval_similarity,
                    include_file=False,
                )
            except colrev_exceptions.RecordNotInTOCException:
                return record

            except colrev_exceptions.RecordNotInIndexException:
                try:
                    # Search across table-of-contents in local_index
                    retrieved_record_dict = self.local_index.retrieve_from_toc(
                        record_dict=record.data,
                        similarity_threshold=retrieval_similarity,
                        include_file=False,
                        search_across_tocs=True,
                    )
                except (
                    colrev_exceptions.RecordNotInIndexException,
                    colrev_exceptions.RecordNotInTOCException,
                ):
                    return record
            except colrev_exceptions.NotTOCIdentifiableException:
                return record
        finally:
            if added_colrev_pdf_id:
                del record.data["colrev_pdf_id"]

        if "colrev_status" in retrieved_record_dict:
            del retrieved_record_dict["colrev_status"]

        return colrev.record.PrepRecord(data=retrieved_record_dict)

    def __store_retrieved_record_in_feed(
        self,
        *,
        record: colrev.record.Record,
        retrieved_record: colrev.record.Record,
        default_source: str,
        prep_operation: colrev.ops.prep.Prep,
    ) -> None:
        try:
            # lock: to prevent different records from having the same origin
            self.local_index_lock.acquire(timeout=60)

            # Note : need to reload file because the object is not shared between processes
            local_index_feed = self.search_source.get_feed(
                review_manager=self.review_manager,
                source_identifier=self.source_identifier,
                update_only=False,
            )

            local_index_feed.set_id(record_dict=retrieved_record.data)
            local_index_feed.add_record(record=retrieved_record)

            retrieved_record.remove_field(key="curation_ID")
            record.merge(
                merging_record=retrieved_record,
                default_source=default_source,
            )
            # If volume/number are no longer in the CURATED record
            if "number" in record.data and "number" not in retrieved_record.data:
                del record.data["number"]
            if "volume" in record.data and "volume" not in retrieved_record.data:
                del record.data["volume"]

            record.set_status(target_state=colrev.record.RecordState.md_prepared)
            if record.data.get("prescreen_exclusion", "NA") == "retracted":
                record.prescreen_exclude(reason="retracted")

            git_repo = self.review_manager.dataset.get_repo()
            cur_project_source_paths = [str(self.review_manager.path)]
            for remote in git_repo.remotes:
                if remote.url:
                    shared_url = remote.url
                    shared_url = shared_url.rstrip(".git")
                    cur_project_source_paths.append(shared_url)
                    break

            try:
                local_index_feed.save_feed_file()
                # extend fields_to_keep (to retrieve all fields from the index)
                for key in record.data.keys():
                    if key not in prep_operation.fields_to_keep:
                        prep_operation.fields_to_keep.append(key)

            except OSError:
                pass

        except (colrev_exceptions.InvalidMerge,):
            print("invalid-merge")
        except colrev_exceptions.NotFeedIdentifiableException:
            print("not-feed-identifiable")
        finally:
            self.local_index_lock.release()

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Retrieve masterdata from LocalIndex based on similarity with the record provided"""

        retrieved_record = self.__retrieve_record_from_local_index(
            record=record,
            retrieval_similarity=prep_operation.retrieval_similarity,
        )

        # restriction: if we don't restrict to CURATED,
        # we may have to rethink the LocalIndexSearchFeed.set_ids()
        if "CURATED" not in retrieved_record.data.get(
            "colrev_masterdata_provenance", ""
        ):
            return record

        default_source = "LOCAL_INDEX"
        if "colrev_masterdata_provenance" in retrieved_record.data:
            if "CURATED" in retrieved_record.data["colrev_masterdata_provenance"]:
                default_source = retrieved_record.data["colrev_masterdata_provenance"][
                    "CURATED"
                ]["source"]

        self.__store_retrieved_record_in_feed(
            record=record,
            retrieved_record=retrieved_record,
            default_source=default_source,
            prep_operation=prep_operation,
        )

        return record

    def __get_local_base_repos(self, *, change_itemsets: list) -> dict:
        base_repos = []
        for item in change_itemsets:
            if "CURATED" in item["original_record"].get(
                "colrev_masterdata_provenance", {}
            ):
                repo_path = item["original_record"]["colrev_masterdata_provenance"][
                    "CURATED"
                ]["source"]
                assert "#" not in repo_path
                # otherwise: strip the ID at the end if we add an ID...
                base_repos.append(repo_path)

        base_repos = list(set(base_repos))
        environment_manager = colrev.env.environment_manager.EnvironmentManager()
        local_base_repos = {
            x["repo_source_url"]: x["repo_source_path"]
            for x in environment_manager.local_repos()
            if x.get("repo_source_url", "NA") in base_repos
        }
        return local_base_repos

    def __print_changes(self, *, local_base_repo: str, change_itemsets: list) -> list:
        def print_diff(change: tuple) -> str:
            diff = difflib.Differ()
            letters = list(diff.compare(change[1], change[0]))
            for i, letter in enumerate(letters):
                if letter.startswith("  "):
                    letters[i] = letters[i][-1]
                elif letter.startswith("+ "):
                    letters[i] = f"{colors.RED}" + letters[i][-1] + f"{colors.END}"
                elif letter.startswith("- "):
                    letters[i] = f"{colors.GREEN}" + letters[i][-1] + f"{colors.END}"
            res = "".join(letters).replace("\n", " ")
            return res

        selected_changes = []
        print()
        self.review_manager.logger.info(f"Base repository: {local_base_repo}")
        for item in change_itemsets:
            repo_path = "NA"
            if "CURATED" in item["original_record"].get(
                "colrev_masterdata_provenance", {}
            ):
                repo_path = item["original_record"]["colrev_masterdata_provenance"][
                    "CURATED"
                ]["source"]
                assert "#" not in repo_path
                # otherwise: strip the ID at the end if we add an ID...

            if repo_path != local_base_repo:
                continue

            # self.review_manager.p_printer.pprint(item["original_record"])
            colrev.record.Record(data=item["original_record"]).print_citation_format()
            for change_item in item["changes"]:
                if change_item[0] == "change":
                    edit_type, field, values = change_item
                    if field == "colrev_id":
                        continue
                    prefix = f"{edit_type} {field}"
                    print(
                        f"{prefix}"
                        + " " * max(len(prefix), 30 - len(prefix))
                        + f": {values[0]}"
                    )
                    print(
                        " " * max(len(prefix), 30)
                        + f"  {colors.ORANGE}{values[1]}{colors.END}"
                    )
                    print(
                        " " * max(len(prefix), 30)
                        + f"  {print_diff((values[0], values[1]))}"
                    )

                elif change_item[0] == "add":
                    edit_type, field, values = change_item
                    prefix = f"{edit_type} {values[0][0]}"
                    print(
                        prefix
                        + " " * max(len(prefix), 30 - len(prefix))
                        + f": {colors.GREEN}{values[0][1]}{colors.END}"
                    )
                else:
                    self.review_manager.p_printer.pprint(change_item)
            selected_changes.append(item)
        return selected_changes

    def apply_correction(self, *, change_itemsets: list) -> None:
        """Apply a correction by opening a pull request in the original repository"""

        local_base_repos = self.__get_local_base_repos(change_itemsets=change_itemsets)

        for local_base_repo_url, local_base_repo_path in local_base_repos.items():
            selected_changes = self.__print_changes(
                local_base_repo=local_base_repo_url, change_itemsets=change_itemsets
            )

            response = ""
            while True:
                response = input("\nConfirm changes? (y/n)")
                if response in ["y", "n"]:
                    break

            if response == "y":
                self.__apply_correction(
                    source_url=local_base_repo_path,
                    change_list=selected_changes,
                )
            elif response == "n":
                if input("Discard all corrections (y/n)?") == "y":
                    for selected_change in selected_changes:
                        Path(selected_change["file"]).unlink()

    def __apply_corrections_precondition(
        self, *, check_operation: colrev.operation.Operation, source_url: str
    ) -> bool:
        git_repo = check_operation.review_manager.dataset.get_repo()

        if git_repo.is_dirty():
            msg = f"Repo not clean ({source_url}): commit or stash before updating records"
            raise colrev_exceptions.CorrectionPreconditionException(msg)

        if check_operation.review_manager.dataset.behind_remote():
            origin = git_repo.remotes.origin
            origin.pull()
            if not check_operation.review_manager.dataset.behind_remote():
                self.review_manager.logger.info("Pulled changes")
            else:
                self.review_manager.logger.error(
                    "Repo behind remote. Pull first to avoid conflicts.\n"
                    "colrev env --pull"
                )
                return False

        return True

    def __retrieve_by_colrev_id(
        self, *, indexed_record_dict: dict, records: list[dict]
    ) -> dict:
        indexed_record = colrev.record.Record(data=indexed_record_dict)

        if "colrev_id" in indexed_record.data:
            cid_to_retrieve = indexed_record.get_colrev_id()
        else:
            cid_to_retrieve = [indexed_record.create_colrev_id()]

        record_l = [
            x
            for x in records
            if any(
                cid in colrev.record.Record(data=x).get_colrev_id()
                for cid in cid_to_retrieve
            )
        ]
        if len(record_l) != 1:
            raise colrev_exceptions.RecordNotInRepoException
        return record_l[0]

    def __retrieve_record_for_correction(
        self,
        *,
        records: dict,
        change_item: dict,
    ) -> dict:
        original_record = change_item["original_record"]

        local_index_feed = self.search_source.get_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=True,
        )

        try:
            md_curated_origin_id = [
                x for x in original_record["colrev_origin"] if "md_curated.bib/" in x
            ][0].replace("md_curated.bib/", "")
            curation_origin_record = local_index_feed.feed_records[md_curated_origin_id]
            curation_id = curation_origin_record["curation_ID"]
            curation_id = curation_id[curation_id.find("#") + 1 :]
            return records[curation_id]
        except KeyError:
            pass

        try:
            record_dict = self.__retrieve_by_colrev_id(
                indexed_record_dict=original_record,
                records=list(records.values()),
            )
            return record_dict
        except colrev_exceptions.RecordNotInRepoException:
            matching_doi_rec_l = [
                r
                for r in records.values()
                if original_record.get("doi", "NDOI") == r.get("doi", "NA")
            ]
            if len(matching_doi_rec_l) == 1:
                record_dict = matching_doi_rec_l[0]
                return record_dict

            matching_url_rec_l = [
                r
                for r in records.values()
                if original_record.get("url", "NURL") == r.get("url", "NA")
            ]
            if len(matching_url_rec_l) == 1:
                record_dict = matching_url_rec_l[0]
                return record_dict

        self.review_manager.logger.error(
            f'{colors.RED}Record not found: {original_record["ID"]}{colors.END}'
        )
        raise colrev_exceptions.RecordNotInIndexException()

    def __create_correction_branch(
        self, *, git_repo: git.Repo, record_dict: dict
    ) -> str:
        record_branch_name = record_dict["ID"]
        counter = 1
        new_record_branch_name = record_branch_name
        while new_record_branch_name in [ref.name for ref in git_repo.references]:
            new_record_branch_name = f"{record_branch_name}_{counter}"
            counter += 1

        record_branch_name = new_record_branch_name
        git_repo.git.branch(record_branch_name)
        return record_branch_name

    def __apply_record_correction(
        self,
        *,
        check_operation: colrev.operation.Operation,
        records: dict,
        record_dict: dict,
        change_item: dict,
    ) -> None:
        for edit_type, key, change in list(change_item["changes"]):
            # Note : by retricting changes to self.essential_md_keys,
            # we also prevent changes in
            # "colrev_status", "colrev_origin", "file"

            # Note: the most important thing is to update the metadata.

            if edit_type == "change":
                if key not in self.essential_md_keys:
                    continue
                record_dict[key] = change[1]
            if edit_type == "add":
                key = change[0][0]
                value = change[0][1]
                if key not in self.essential_md_keys:
                    continue
                record_dict[key] = value
            # gh_issue https://github.com/CoLRev-Environment/colrev/issues/63
            # deal with remove/merge

        check_operation.review_manager.dataset.save_records_dict(records=records)
        check_operation.review_manager.dataset.add_record_changes()
        check_operation.review_manager.create_commit(
            msg=f"Update {record_dict['ID']}", script_call="colrev push"
        )

    def __push_corrections_and_reset_branch(
        self,
        *,
        git_repo: git.Repo,
        record_branch_name: str,
        prev_branch_name: str,
        source_url: str,
    ) -> None:
        git_repo.remotes.origin.push(
            refspec=f"{record_branch_name}:{record_branch_name}"
        )
        self.review_manager.logger.info("Pushed corrections")

        for head in git_repo.heads:
            if head.name == prev_branch_name:
                head.checkout()

        git_repo = git.Git(source_url)
        git_repo.execute(["git", "branch", "-D", record_branch_name])

        self.review_manager.logger.info("Removed local corrections branch")

    def __reset_record_after_correction(
        self, *, record_dict: dict, rec_for_reset: dict, change_item: dict
    ) -> None:
        # reset the record - each branch should have changes for one record
        # Note : modify dict (do not replace it) - otherwise changes will not be
        # part of the records.
        for key, value in rec_for_reset.items():
            record_dict[key] = value
        keys_added = [
            key for key in record_dict.keys() if key not in rec_for_reset.keys()
        ]
        for key in keys_added:
            del record_dict[key]

        if Path(change_item["file"]).is_file():
            Path(change_item["file"]).unlink()

    def __apply_change_item_correction(
        self,
        *,
        check_operation: colrev.operation.Operation,
        source_url: str,
        change_list: list,
    ) -> bool:
        # pylint: disable=too-many-locals

        git_repo = check_operation.review_manager.dataset.get_repo()
        records = check_operation.review_manager.dataset.load_records_dict()

        success = False
        pull_request_msgs = []
        pull_request_links = []
        for change_item in change_list:
            try:
                record_dict = self.__retrieve_record_for_correction(
                    records=records,
                    change_item=change_item,
                )
                if not record_dict:
                    continue

                record_branch_name = self.__create_correction_branch(
                    git_repo=git_repo, record_dict=record_dict
                )
                prev_branch_name = git_repo.active_branch.name

                remote = git_repo.remote()
                for head in git_repo.heads:
                    if head.name == record_branch_name:
                        head.checkout()

                rec_for_reset = record_dict.copy()

                self.__apply_record_correction(
                    check_operation=check_operation,
                    records=records,
                    record_dict=record_dict,
                    change_item=change_item,
                )

                self.__push_corrections_and_reset_branch(
                    git_repo=git_repo,
                    record_branch_name=record_branch_name,
                    prev_branch_name=prev_branch_name,
                    source_url=source_url,
                )

                self.__reset_record_after_correction(
                    record_dict=record_dict,
                    rec_for_reset=rec_for_reset,
                    change_item=change_item,
                )

                host = urlparse(remote.url).hostname
                if host and host.endswith("github.com"):
                    link = (
                        str(remote.url).rstrip(".git")
                        + "/compare/"
                        + record_branch_name
                    )
                    pull_request_links.append(link)
                    pull_request_msgs.append(
                        "\nTo create a pull request for your changes go "
                        f"to \n{colors.ORANGE}{link}{colors.END}"
                    )
                success = True
            except colrev_exceptions.RecordNotInIndexException:
                pass

        for pull_request_msg in pull_request_msgs:
            print(pull_request_msg)
        for pull_request_link in pull_request_links:
            webbrowser.open(pull_request_link, new=2)

        # https://github.com/geritwagner/information_systems_papers/compare/update?expand=1
        # gh_issue https://github.com/CoLRev-Environment/colrev/issues/63
        # handle cases where update branch already exists
        return success

    def __apply_correction(self, *, source_url: str, change_list: list) -> None:
        """Apply a (list of) corrections"""

        # TBD: other modes of accepting changes?
        # e.g., only-metadata, no-changes, all(including optional fields)
        check_review_manager = self.review_manager.get_connecting_review_manager(
            path_str=source_url
        )
        check_operation = colrev.operation.CheckOperation(
            review_manager=check_review_manager
        )

        if check_review_manager.dataset.behind_remote():
            git_repo = check_review_manager.dataset.get_repo()
            origin = git_repo.remotes.origin
            self.review_manager.logger.info(
                f"Pull project changes from {git_repo.remotes.origin}"
            )
            res = origin.pull()
            self.review_manager.logger.info(res)

        try:
            if not self.__apply_corrections_precondition(
                check_operation=check_operation, source_url=source_url
            ):
                return
        except colrev_exceptions.CorrectionPreconditionException as exc:
            print(exc)
            return

        check_review_manager.logger.info(
            "Precondition for correction (pull-request) checked."
        )

        success = self.__apply_change_item_correction(
            check_operation=check_operation,
            source_url=source_url,
            change_list=change_list,
        )

        if success:
            print(
                f"\n{colors.GREEN}Thank you for supporting other researchers "
                f"by sharing your corrections ❤{colors.END}\n"
            )



#! /usr/bin/env python
"""Utility for prep"""
from __future__ import annotations

import re

NO_CAPS = ["of", "for", "the", "and"]
ALL_CAPS = ["IEEE", "ACM", "M&A", "B2B", "B2C", "C2C", "I"]
ALL_CAPS_DICT = {r"U\.S\.": "U.S."}


def capitalize_entities(input_str: str) -> str:
    """Utility function to capitalize entities"""

    for all_cap in ALL_CAPS:
        input_str = re.sub(
            rf"\b{all_cap.lower()}\b", all_cap.upper(), input_str, flags=re.IGNORECASE
        )

    for all_cap, repl in ALL_CAPS_DICT.items():
        input_str = re.sub(
            rf"\b{all_cap.lower()}\b", repl, input_str, flags=re.IGNORECASE
        )

    for no_cap in NO_CAPS:
        if input_str.lower().startswith(no_cap):
            continue
        input_str = re.sub(
            rf"\b{no_cap.lower()}\b", no_cap, input_str, flags=re.IGNORECASE
        )

    input_str = input_str.replace(" i'", " I'").replace("'S ", "'s ")

    input_str = re.sub(r"it-(\w)", r"IT-\1", input_str, flags=re.IGNORECASE)
    input_str = re.sub(r"is-(\w)", r"IS-\1", input_str, flags=re.IGNORECASE)

    return input_str



#! /usr/bin/env python
"""Simple dedupe functionality (based on similarity thresholds) for small samples"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import pandas as pd
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.built_in.dedupe.utils
import colrev.record
import colrev.ui_cli.cli_colors as colors

if TYPE_CHECKING:
    import colrev.ops.dedupe

# pylint: disable=too-many-arguments
# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.DedupePackageEndpointInterface)
@dataclass
class SimpleDedupe(JsonSchemaMixin):
    """Simple duplicate identification (for small sample sizes)"""

    settings: SimpleDedupeSettings

    ci_supported: bool = True

    @dataclass
    class SimpleDedupeSettings(
        colrev.env.package_manager.DefaultSettings, JsonSchemaMixin
    ):
        """Settings for SimpleDedupe"""

        endpoint: str
        merging_non_dup_threshold: float = 0.7
        merging_dup_threshold: float = 0.95

        _details = {
            "merging_non_dup_threshold": {
                "tooltip": "Threshold: record pairs with a similarity "
                "below this threshold are considered non-duplicates"
            },
            "merging_dup_threshold": {
                "tooltip": "Threshold: record pairs with a similarity "
                "above this threshold are considered duplicates"
            },
        }

    settings_class = SimpleDedupeSettings

    def __init__(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,  # pylint: disable=unused-argument
        settings: dict,
    ):
        self.settings = self.settings_class.load_settings(data=settings)

        assert self.settings.merging_non_dup_threshold >= 0.0
        assert self.settings.merging_non_dup_threshold <= 1.0
        assert self.settings.merging_dup_threshold >= 0.0
        assert self.settings.merging_dup_threshold <= 1.0

    def __get_maximum_similarity_record(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,  # pylint: disable=unused-argument
        records_batch: list,
    ) -> dict:
        max_similarity_record = {
            "reference_record": records_batch[len(records_batch) - 1]["ID"],
            "record_id": "NA",
            "similarity": 0,
        }
        for i in range(0, len(records_batch) - 1):
            sim_details = colrev.record.Record.get_similarity_detailed(
                record_a=records_batch[i],
                record_b=records_batch[len(records_batch) - 1],
            )

            if sim_details["score"] > max_similarity_record["similarity"]:
                max_similarity_record["similarity"] = sim_details["score"]
                max_similarity_record["record_id"] = records_batch[i]["ID"]
                max_similarity_record["details"] = sim_details["details"]

        return max_similarity_record

    def __append_merges(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe, batch_item: dict
    ) -> dict:
        records_batch = batch_item["queue"]

        # if the record is the first one added to the records
        # (in a preceding processing step), it can be propagated
        if len(records_batch) < 2:
            return {
                "ID1": batch_item["record"],
                "ID2": "NA",
                "similarity": 1,
                "decision": "no_duplicate",
            }

        # df to get_similarities for each other record
        similarity_dict = self.__get_maximum_similarity_record(
            dedupe_operation=dedupe_operation, records_batch=records_batch
        )

        max_similarity = similarity_dict["similarity"]

        ret = {}
        if max_similarity <= self.settings.merging_non_dup_threshold:
            # Note: if no other record has a similarity exceeding the threshold,
            # it is considered a non-duplicate (in relation to all other records)
            # dedupe_operation.review_manager.logger.debug(
            #     f"max_similarity ({max_similarity})"
            # )
            ret = {
                "ID1": similarity_dict["reference_record"],
                "ID2": "NA",
                "similarity": max_similarity,
                "decision": "no_duplicate",
            }

        elif (
            self.settings.merging_non_dup_threshold
            < max_similarity
            < self.settings.merging_dup_threshold
        ):
            other_id = similarity_dict["record_id"]
            # dedupe_operation.review_manager.logger.debug(
            #     f"max_similarity ({max_similarity}): {batch_item['record']} {other_id}"
            # )
            # details = similarity_dict["details"]
            # dedupe_operation.review_manager.logger.debug(details)
            # record_a, record_b = sorted([ID, record["ID"]])
            msg = (
                f'{similarity_dict["reference_record"]} - {other_id}'.ljust(35, " ")
                + f"  - potential duplicate (similarity: {max_similarity})"
            )
            # dedupe_operation.review_manager.report_logger.info(msg)
            dedupe_operation.review_manager.logger.info(msg)
            ret = {
                "ID1": similarity_dict["reference_record"],
                "ID2": other_id,
                "similarity": max_similarity,
                "decision": "potential_duplicate",
            }

        else:  # max_similarity >= self.settings.merging_dup_threshold:
            # note: the following status will not be saved in the bib file but
            # in the duplicate_tuples.csv (which will be applied to the bib file
            # in the end)
            other_id = similarity_dict["record_id"]

            # dedupe_operation.review_manager.logger.debug(
            #     f"max_similarity ({max_similarity}): {batch_item['record']} {other_id}"
            # )
            # details = similarity_dict["details"]

            # dedupe_operation.review_manager.logger.debug(details)
            msg = (
                "Dropped duplicate: "
                f'{similarity_dict["reference_record"]} (duplicate of {other_id})'
                # + f" (similarity: {max_similarity})\nDetails: {details}"
            )
            dedupe_operation.review_manager.report_logger.info(msg)
            dedupe_operation.review_manager.logger.info(msg)
            ret = {
                "ID1": similarity_dict["reference_record"],
                "ID2": other_id,
                "similarity": max_similarity,
                "decision": "duplicate",
            }
        return ret

    def __get_dedupe_data(self, *, dedupe_operation: colrev.ops.dedupe.Dedupe) -> dict:
        records_headers = dedupe_operation.review_manager.dataset.load_records_dict(
            header_only=True
        )
        record_header_list = list(records_headers.values())

        ids_to_dedupe = [
            x["ID"]
            for x in record_header_list
            if x["colrev_status"] == colrev.record.RecordState.md_prepared
        ]
        processed_ids = [
            x["ID"]
            for x in record_header_list
            if x["colrev_status"]
            not in [
                colrev.record.RecordState.md_imported,
                colrev.record.RecordState.md_prepared,
                colrev.record.RecordState.md_needs_manual_preparation,
            ]
        ]
        if len(ids_to_dedupe) > 40:
            if not dedupe_operation.review_manager.force_mode:
                dedupe_operation.review_manager.logger.warning(
                    "Simple duplicate identification selected despite sufficient sample size.\n"
                    "Active learning algorithms may perform better:\n"
                    f"{colors.ORANGE}   colrev settings -m 'dedupe.dedupe_package_endpoints="
                    '[{"endpoint": "colrev.active_learning_training"},'
                    f'{{"endpoint": "colrev.active_learning_automated"}}]\'{colors.END}'
                )
                raise colrev_exceptions.CoLRevException(
                    "To use simple duplicate identification, use\n"
                    f"{colors.ORANGE}    colrev dedupe --force{colors.END}"
                )

        nr_tasks = len(ids_to_dedupe)
        dedupe_data = {
            "nr_tasks": nr_tasks,
            "queue": processed_ids + ids_to_dedupe,
            "items_start": len(processed_ids),
        }
        # dedupe_operation.review_manager.logger.debug(
        #     dedupe_operation.review_manager.p_printer.pformat(dedupe_data)
        # )
        return dedupe_data

    def __get_record_batch(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe, dedupe_data: dict
    ) -> list:
        records = dedupe_operation.review_manager.dataset.load_records_dict()

        # Note: Because we only introduce individual (non-merged records),
        # there should be no semicolons in colrev_origin!
        records_queue = [
            record
            for ID, record in records.items()
            if ID in dedupe_data["queue"]  # type: ignore
        ]

        records_df_queue = pd.DataFrame.from_records(records_queue)
        records = dedupe_operation.prep_records(records_df=records_df_queue)
        # dedupe.review_manager.p_printer.pprint(records.values())

        items_start = dedupe_data["items_start"]
        items_start = 0
        batch_data = []
        for i in range(items_start, len(dedupe_data["queue"])):  # type: ignore
            batch_data.append(
                {
                    "record": dedupe_data["queue"][i],  # type: ignore
                    "queue": list(records.values())[: i + 1],
                }
            )
        return batch_data

    def __process_potential_duplicates(
        self, *, dedupe_operation: colrev.ops.dedupe.Dedupe, dedupe_batch_results: list
    ) -> list:
        potential_duplicates = [
            r for r in dedupe_batch_results if "potential_duplicate" == r["decision"]
        ]

        records = dedupe_operation.review_manager.dataset.load_records_dict()
        records = dedupe_operation.prep_records(
            records_df=pd.DataFrame.from_records(list(records.values()))
        )
        # dedupe.review_manager.p_printer.pprint(records.values())
        records_df = pd.DataFrame(records.values())

        keys = list(records_df.columns)
        for key_to_drop in [
            "ID",
            "colrev_origin",
            "colrev_status",
            "colrev_id",
            "container_title",
        ]:
            if key_to_drop in keys:
                keys.remove(key_to_drop)

        n_match, n_distinct = 0, 0
        for potential_duplicate in potential_duplicates:
            rec1 = records_df.loc[records_df["ID"] == potential_duplicate["ID1"], :]
            rec2 = records_df.loc[records_df["ID"] == potential_duplicate["ID2"], :]

            record_pair = [rec1.to_dict("records")[0], rec2.to_dict("records")[0]]

            user_input = (
                colrev.ops.built_in.dedupe.utils.console_duplicate_instance_label(
                    record_pair, keys, True, "TODO", n_match, n_distinct, []
                )
            )

            # set potential_duplicates
            if user_input == "y":
                potential_duplicate["decision"] = "duplicate"
                n_match += 1
            if user_input == "n":
                potential_duplicate["decision"] = "no_duplicate"
                n_distinct += 1

        return potential_duplicates

    def run_dedupe(self, dedupe_operation: colrev.ops.dedupe.Dedupe) -> None:
        """Pairwise identification of duplicates based on static similarity measure

        This procedure should only be used in small samples on which active learning
        models cannot be trained.
        """

        # default='warn'
        pd.options.mode.chained_assignment = None  # type: ignore  # noqa

        dedupe_operation.review_manager.logger.info(
            "Dedupe operation [colrev.simple_dedupe]"
        )
        dedupe_operation.review_manager.logger.info(
            "Duplicate identification based on static similarity measure and record pairs"
        )

        dedupe_data = self.__get_dedupe_data(dedupe_operation=dedupe_operation)

        # the queue (order) matters for the incremental merging (make sure that each
        # additional record is compared to/merged with all prior records in
        # the queue)

        if not dedupe_data["queue"]:
            dedupe_operation.review_manager.logger.error("No records to dedupe")
            return

        batch_data = self.__get_record_batch(
            dedupe_operation=dedupe_operation, dedupe_data=dedupe_data
        )

        dedupe_batch_results = []
        for item in batch_data:
            merge_item = self.__append_merges(
                dedupe_operation=dedupe_operation, batch_item=item
            )
            dedupe_batch_results.append(merge_item)

        # dedupe_batch[-1]['queue'].to_csv('last_records.csv')

        dedupe_operation.apply_merges(
            results=dedupe_batch_results, complete_dedupe=True
        )
        if [x for x in dedupe_batch_results if x["decision"] == "duplicate"]:
            dedupe_operation.review_manager.logger.info(
                "Completed application of merges"
            )

        dedupe_operation.review_manager.create_commit(
            msg="Merge duplicate records",
        )

        dedupe_operation.review_manager.logger.info(
            "Potential duplicate identification based on static similarity measure and record pairs"
        )
        potential_duplicates = self.__process_potential_duplicates(
            dedupe_operation=dedupe_operation, dedupe_batch_results=dedupe_batch_results
        )

        # apply:
        dedupe_operation.apply_merges(results=potential_duplicates)

        # add and commit
        dedupe_operation.review_manager.dataset.add_record_changes()
        dedupe_operation.review_manager.create_commit(
            msg="Manual labeling of remaining duplicate candidates",
            manual_author=False,
        )



#! /usr/bin/env python
"""SearchSource: Web of Science"""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.load_utils_bib
import colrev.ops.search
import colrev.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class WebOfScienceSearchSource(JsonSchemaMixin):
    """SearchSource for Web of Science"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = (
        "https://www.webofscience.com/wos/woscc/full-record/" + "{{unique-id}}"
    )
    search_type = colrev.settings.SearchType.DB
    api_search_supported = False
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "Web of Science"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/web_of_science.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.Operation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Web of Science"""

        result = {"confidence": 0.0}

        if data.count("UT WOS:") > 0.4 * data.count("TI "):
            result["confidence"] = 0.7
            return result

        if "Unique-ID = {WOS:" in data:
            result["confidence"] = 0.7
            return result
        if "UT_(Unique_WOS_ID) = {WOS:" in data:
            result["confidence"] = 0.7
            return result
        if "@article{ WOS:" in data or "@article{WOS:" in data:
            if data.count("{WOS:") > data.count("\n@"):
                result["confidence"] = 1.0
            elif data.count("{ WOS:") > data.count("\n@"):
                result["confidence"] = 1.0
            else:
                result["confidence"] = 0.7

            return result

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""
        raise NotImplementedError

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of WebOfScience"""

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.PrepRecord, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for Web of Science"""

        record.format_if_mostly_upper(key="title", case="sentence")
        record.format_if_mostly_upper(key="journal", case="title")
        record.format_if_mostly_upper(key="booktitle", case="title")
        record.format_if_mostly_upper(key="author", case="title")

        record.remove_field(key="colrev.web_of_science.researcherid-numbers")
        record.remove_field(key="colrev.web_of_science.orcid-numbers")

        return record



#!/usr/bin/env python3
"""Upgrades CoLRev projects."""
from __future__ import annotations

import json
import shutil
import typing
from importlib.metadata import version
from pathlib import Path
from typing import TYPE_CHECKING

import git
from tqdm import tqdm

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.operation
import colrev.ui_cli.cli_colors as colors

if TYPE_CHECKING:
    import colrev.review_manager


# pylint: disable=too-few-public-methods


class Upgrade(colrev.operation.Operation):
    """Upgrade a CoLRev project"""

    repo: git.Repo

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
    ) -> None:
        prev_force_mode = review_manager.force_mode
        review_manager.force_mode = True
        super().__init__(
            review_manager=review_manager,
            operations_type=colrev.operation.OperationsType.check,
            notify_state_transition_operation=False,
        )
        review_manager.force_mode = prev_force_mode
        self.review_manager = review_manager

    def __move_file(self, source: Path, target: Path) -> None:
        target.parent.mkdir(exist_ok=True, parents=True)
        if source.is_file():
            shutil.move(str(source), self.review_manager.path / target)
            self.repo.index.remove([str(source)])
            self.repo.index.add([str(target)])

    def __load_settings_dict(self) -> dict:
        if not self.review_manager.settings_path.is_file():
            raise colrev_exceptions.CoLRevException()
        with open(self.review_manager.settings_path, encoding="utf-8") as file:
            settings = json.load(file)
        return settings

    def __save_settings(self, settings: dict) -> None:
        with open("settings.json", "w", encoding="utf-8") as outfile:
            json.dump(settings, outfile, indent=4)
        self.repo.index.add(["settings.json"])

    def main(self) -> None:
        """Upgrade a CoLRev project (main entrypoint)"""

        try:
            self.repo = git.Repo(str(self.review_manager.path))
            self.repo.iter_commits()
        except ValueError:
            # Git repository has no initial commit
            return

        settings = self.__load_settings_dict()
        settings_version_str = settings["project"]["colrev_version"]

        settings_version = CoLRevVersion(settings_version_str)
        # Start with the first step if the version is older:
        if settings_version < CoLRevVersion("0.7.0"):
            settings_version = CoLRevVersion("0.7.0")
        installed_colrev_version = CoLRevVersion(version("colrev"))

        # version: indicates from which version on the migration should be applied
        migration_scripts: typing.List[typing.Dict[str, typing.Any]] = [
            {
                "version": CoLRevVersion("0.7.0"),
                "target_version": CoLRevVersion("0.7.1"),
                "script": self.__migrate_0_7_0,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.7.1"),
                "target_version": CoLRevVersion("0.8.0"),
                "script": self.__migrate_0_7_1,
                "released": True,
            },
            # Note : we may add a flag to update to pre-released versions
            {
                "version": CoLRevVersion("0.8.0"),
                "target_version": CoLRevVersion("0.8.1"),
                "script": self.__migrate_0_8_0,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.8.1"),
                "target_version": CoLRevVersion("0.8.2"),
                "script": self.__migrate_0_8_1,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.8.2"),
                "target_version": CoLRevVersion("0.8.3"),
                "script": self.__migrate_0_8_2,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.8.3"),
                "target_version": CoLRevVersion("0.8.4"),
                "script": self.__migrate_0_8_3,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.8.4"),
                "target_version": CoLRevVersion("0.9.0"),
                "script": self.__migrate_0_8_4,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.9.0"),
                "target_version": CoLRevVersion("0.9.1"),
                "script": self.__migrate_0_9_1,
                "released": True,
            },
            {
                "version": CoLRevVersion("0.9.2"),
                "target_version": CoLRevVersion("0.9.3"),
                "script": self.__migrate_0_9_3,
                "released": True,
            },
        ]

        # Note: we should always update the colrev_version in settings.json because the
        # checker.__check_software requires the settings version and
        # the installed version to be identical

        # skipping_versions_before_settings_version = True
        run_migration = False
        while migration_scripts:
            migrator = migration_scripts.pop(0)
            # Activate run_migration for the current settings_version
            if (
                settings_version == migrator["version"]
            ):  # settings_version == migrator["version"] or
                run_migration = True
            if not run_migration:
                continue

            if installed_colrev_version == settings_version and migrator["released"]:
                return

            migration_script = migrator["script"]
            self.review_manager.logger.info(
                "Upgrade to: %s", migrator["target_version"]
            )
            if migrator["released"]:
                self.__print_release_notes(selected_version=migrator["target_version"])

            updated = migration_script()
            if not updated:
                continue

        settings = self.__load_settings_dict()
        settings["project"]["colrev_version"] = str(installed_colrev_version)
        self.__save_settings(settings)

        if self.repo.is_dirty():
            msg = f"Upgrade to CoLRev {installed_colrev_version}"
            if not migrator["released"]:
                msg += " (pre-release)"
            review_manager = colrev.review_manager.ReviewManager()
            review_manager.create_commit(
                msg=msg,
            )

    def __print_release_notes(self, *, selected_version: CoLRevVersion) -> None:
        filedata = colrev.env.utils.get_package_file_content(
            file_path=Path("../CHANGELOG.md")
        )
        active, printed = False, False
        if filedata:
            for line in filedata.decode("utf-8").split("\n"):
                if str(selected_version) in line:
                    active = True
                    print(f"{colors.ORANGE}Release notes v{selected_version}")
                    continue
                if line.startswith("## "):
                    active = False
                if active:
                    print(line)
                    printed = True
        if not printed:
            print(f"{colors.ORANGE}No release notes")
        print(f"{colors.END}")

    def __migrate_0_7_0(self) -> bool:
        pre_commit_contents = Path(".pre-commit-config.yaml").read_text(
            encoding="utf-8"
        )
        if "ci:" not in pre_commit_contents:
            pre_commit_contents = pre_commit_contents.replace(
                "repos:",
                "ci:\n    skip: [colrev-hooks-format, colrev-hooks-check]\n\nrepos:",
            )
            with open(".pre-commit-config.yaml", "w", encoding="utf-8") as file:
                file.write(pre_commit_contents)
        self.repo.index.add([".pre-commit-config.yaml"])
        return self.repo.is_dirty()

    def __migrate_0_7_1(self) -> bool:
        settings_content = (self.review_manager.path / Path("settings.json")).read_text(
            encoding="utf-8"
        )
        settings_content = settings_content.replace("colrev_built_in.", "colrev.")

        with open(Path("settings.json"), "w", encoding="utf-8") as file:
            file.write(settings_content)

        self.repo.index.add(["settings.json"])
        self.review_manager.load_settings()
        if self.review_manager.settings.is_curated_masterdata_repo():
            self.review_manager.settings.project.delay_automated_processing = False
        self.review_manager.save_settings()

        self.__move_file(
            source=Path("data/paper.md"), target=Path("data/data/paper.md")
        )
        self.__move_file(
            source=Path("data/APA-7.docx"), target=Path("data/data/APA-7.docx")
        )
        self.__move_file(
            source=Path("data/non_sample_references.bib"),
            target=Path("data/data/non_sample_references.bib"),
        )

        return self.repo.is_dirty()

    def __migrate_0_8_0(self) -> bool:
        Path(".github/workflows/").mkdir(exist_ok=True, parents=True)

        if "colrev/curated_metadata" in str(self.review_manager.path):
            Path(".github/workflows/colrev_update.yml").unlink(missing_ok=True)
            colrev.env.utils.retrieve_package_file(
                template_file=Path("template/init/colrev_update_curation.yml"),
                target=Path(".github/workflows/colrev_update.yml"),
            )
            self.repo.index.add([".github/workflows/colrev_update.yml"])
        else:
            Path(".github/workflows/colrev_update.yml").unlink(missing_ok=True)
            colrev.env.utils.retrieve_package_file(
                template_file=Path("template/init/colrev_update.yml"),
                target=Path(".github/workflows/colrev_update.yml"),
            )
            self.repo.index.add([".github/workflows/colrev_update.yml"])

        Path(".github/workflows/pre-commit.yml").unlink(missing_ok=True)
        colrev.env.utils.retrieve_package_file(
            template_file=Path("template/init/pre-commit.yml"),
            target=Path(".github/workflows/pre-commit.yml"),
        )
        self.repo.index.add([".github/workflows/pre-commit.yml"])
        return self.repo.is_dirty()

    def __migrate_0_8_1(self) -> bool:
        Path(".github/workflows/").mkdir(exist_ok=True, parents=True)
        if "colrev/curated_metadata" in str(self.review_manager.path):
            Path(".github/workflows/colrev_update.yml").unlink(missing_ok=True)
            colrev.env.utils.retrieve_package_file(
                template_file=Path("template/init/colrev_update_curation.yml"),
                target=Path(".github/workflows/colrev_update.yml"),
            )
            self.repo.index.add([".github/workflows/colrev_update.yml"])
        else:
            Path(".github/workflows/colrev_update.yml").unlink(missing_ok=True)
            colrev.env.utils.retrieve_package_file(
                template_file=Path("template/init/colrev_update.yml"),
                target=Path(".github/workflows/colrev_update.yml"),
            )
            self.repo.index.add([".github/workflows/colrev_update.yml"])

        settings = self.__load_settings_dict()
        settings["project"]["auto_upgrade"] = True
        self.__save_settings(settings)

        return self.repo.is_dirty()

    def __migrate_0_8_2(self) -> bool:
        records = self.review_manager.dataset.load_records_dict()

        for record_dict in tqdm(records.values()):
            if "colrev_pdf_id" not in record_dict:
                continue
            if not record_dict["colrev_pdf_id"].startswith("cpid1:"):
                continue
            if not Path(record_dict.get("file", "")).is_file():
                continue

            pdf_path = Path(record_dict["file"])
            colrev_pdf_id = colrev.record.Record.get_colrev_pdf_id(pdf_path=pdf_path)
            record_dict["colrev_pdf_id"] = colrev_pdf_id

        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()

        return self.repo.is_dirty()

    def __migrate_0_8_3(self) -> bool:
        # pylint: disable=too-many-branches
        settings = self.__load_settings_dict()
        settings["prep"]["defects_to_ignore"] = []
        if "curated_metadata" in str(self.review_manager.path):
            settings["prep"]["defects_to_ignore"] = [
                "record-not-in-toc",
                "inconsistent-with-url-metadata",
            ]
        else:
            settings["prep"]["defects_to_ignore"] = ["inconsistent-with-url-metadata"]

        for p_round in settings["prep"]["prep_rounds"]:
            p_round["prep_package_endpoints"] = [
                x
                for x in p_round["prep_package_endpoints"]
                if x["endpoint"] != "colrev.global_ids_consistency_check"
            ]
        self.__save_settings(settings)
        self.review_manager = colrev.review_manager.ReviewManager(
            path_str=str(self.review_manager.path), force_mode=True
        )
        self.review_manager.load_settings()
        self.review_manager.get_load_operation()
        records = self.review_manager.dataset.load_records_dict()
        quality_model = self.review_manager.get_qm()

        # delete the masterdata provenance notes and apply the new quality model
        # replace not_missing > not-missing
        for record_dict in tqdm(records.values()):
            if "colrev_masterdata_provenance" not in record_dict:
                continue
            not_missing_fields = []
            for key, prov in record_dict["colrev_masterdata_provenance"].items():
                if "not_missing" in prov["note"]:
                    not_missing_fields.append(key)
                prov["note"] = ""
            for key in not_missing_fields:
                record_dict["colrev_masterdata_provenance"][key]["note"] = "not-missing"
            if "cited_by_file" in record_dict:
                del record_dict["cited_by_file"]
            if "cited_by_id" in record_dict:
                del record_dict["cited_by_id"]
            if "tei_id" in record_dict:
                del record_dict["tei_id"]
            if "colrev_data_provenance" in record_dict:
                if "cited_by_file" in record_dict["colrev_data_provenance"]:
                    del record_dict["colrev_data_provenance"]["cited_by_file"]
                if "cited_by_id" in record_dict["colrev_data_provenance"]:
                    del record_dict["colrev_data_provenance"]["cited_by_id"]
                if "tei_id" in record_dict["colrev_data_provenance"]:
                    del record_dict["colrev_data_provenance"]["tei_id"]

            record = colrev.record.Record(data=record_dict)
            prior_state = record.data["colrev_status"]
            record.update_masterdata_provenance(qm=quality_model)
            if prior_state == colrev.record.RecordState.rev_prescreen_excluded:
                record.data[  # pylint: disable=direct-status-assign
                    "colrev_status"
                ] = colrev.record.RecordState.rev_prescreen_excluded
        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()
        return self.repo.is_dirty()

    def __migrate_0_8_4(self) -> bool:
        records = self.review_manager.dataset.load_records_dict()
        for record in records.values():
            if "editor" not in record.get("colrev_data_provenance", {}):
                continue
            ed_val = record["colrev_data_provenance"]["editor"]
            del record["colrev_data_provenance"]["editor"]
            if "CURATED" not in record["colrev_masterdata_provenance"]:
                record["colrev_masterdata_provenance"]["editor"] = ed_val

        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()

        return self.repo.is_dirty()

    def __migrate_0_9_1(self) -> bool:
        settings = self.__load_settings_dict()
        for source in settings["sources"]:
            if "load_conversion_package_endpoint" in source:
                del source["load_conversion_package_endpoint"]
        self.__save_settings(settings)
        return self.repo.is_dirty()

    def __migrate_0_9_3(self) -> bool:
        settings = self.__load_settings_dict()
        for source in settings["sources"]:
            if source["endpoint"] == "colrev.crossref":
                if "issn" not in source["search_parameters"].get("scope", {}):
                    continue
                if isinstance(source["search_parameters"]["scope"]["issn"], str):
                    source["search_parameters"]["scope"]["issn"] = [
                        source["search_parameters"]["scope"]["issn"]
                    ]

        self.__save_settings(settings)

        records = self.review_manager.dataset.load_records_dict()
        for record_dict in records.values():
            # TODO : call methods in repare.py

            if "pubmedid" in record_dict:
                record = colrev.record.Record(data=record_dict)
                record.rename_field(key="pubmedid", new_key="colrev.pubmed.pubmedid")

            if "pii" in record_dict:
                record = colrev.record.Record(data=record_dict)
                record.rename_field(key="pii", new_key="colrev.pubmed.pii")

            if "pmc" in record_dict:
                record = colrev.record.Record(data=record_dict)
                record.rename_field(key="pmc", new_key="colrev.pubmed.pmc")

            if "label_included" in record_dict:
                record = colrev.record.Record(data=record_dict)
                record.rename_field(
                    key="label_included",
                    new_key="colrev.synergy_datasets.label_included",
                )
            if "method" in record_dict:
                record = colrev.record.Record(data=record_dict)
                record.rename_field(
                    key="method", new_key="colrev.synergy_datasets.method"
                )

            if "dblp_key" in record_dict:
                record = colrev.record.Record(data=record_dict)
                record.rename_field(key="dblp_key", new_key="colrev.dblp.dblp_key")
            if "wos_accession_number" in record_dict:
                record = colrev.record.Record(data=record_dict)
                record.rename_field(
                    key="wos_accession_number",
                    new_key="colrev.web_of_science.unique-id",
                )
            if "sem_scholar_id" in record_dict:
                record = colrev.record.Record(data=record_dict)
                record.rename_field(
                    key="sem_scholar_id", new_key="colrev.semantic_scholar.id"
                )

            # TODO : update open_alex source
            if "openalex_id" in record_dict:
                record = colrev.record.Record(data=record_dict)
                record.rename_field(key="openalex_id", new_key="colrev.open_alex.id")

        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()
        return self.repo.is_dirty()


# Note: we can ask users to make decisions (when defaults are not clear)
# via input() or simply cancel the process (raise a CoLrevException)


class CoLRevVersion:
    """Class for handling the CoLRev version"""

    def __init__(self, version_string: str) -> None:
        if "+" in version_string:
            version_string = version_string[: version_string.find("+")]

        self.major = version_string[: version_string.find(".")]
        self.minor = version_string[
            version_string.find(".") + 1 : version_string.rfind(".")
        ]
        self.patch = version_string[version_string.rfind(".") + 1 :]

    def __eq__(self, other) -> bool:  # type: ignore
        return str(self) == str(other)

    def __lt__(self, other) -> bool:  # type: ignore
        if self.major < other.major:
            return True
        if self.minor < other.minor:
            return True
        if self.patch < other.patch:
            return True
        return False

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"



#! /usr/bin/env python
"""Checker for incomplete fields."""
from __future__ import annotations

import colrev.qm.quality_model

# pylint: disable=too-few-public-methods


class IncompleteFieldChecker:
    """The IncompleteFieldChecker"""

    msg = "incomplete-field"

    def __init__(self, quality_model: colrev.qm.quality_model.QualityModel) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the missing-field checks"""

        for key in ["title", "journal", "booktitle", "author", "abstract"]:
            if record.data.get(key, "UNKNOWN") == "UNKNOWN":
                continue
            if self.__incomplete_field(record=record, key=key):
                record.add_masterdata_provenance_note(key=key, note=self.msg)
            else:
                record.remove_masterdata_provenance_note(key=key, note=self.msg)

    def __incomplete_field(self, *, record: colrev.record.Record, key: str) -> bool:
        """check for incomplete field"""
        if record.data[key].endswith("...") or record.data[key].endswith("…"):
            return True
        return key == "author" and (
            # heuristics for missing first names:
            ", and " in record.data[key]
            or record.data[key].rstrip().endswith(",")
            or "," not in record.data[key]
        )


def register(quality_model: colrev.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(IncompleteFieldChecker(quality_model))



#! /usr/bin/env python
"""Export of bib/pdfs as a prep-man operation"""
from __future__ import annotations

import platform
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin
from PyPDF2 import PdfFileReader
from PyPDF2 import PdfFileWriter

import colrev.env.package_manager
import colrev.env.utils
import colrev.record
import colrev.ui_cli.cli_colors as colors

# pylint: disable=duplicate-code
# pylint: disable=too-few-public-methods

if TYPE_CHECKING:
    import colrev.ops.prep_man


@zope.interface.implementer(colrev.env.package_manager.PrepManPackageEndpointInterface)
@dataclass
class ExportManPrep(JsonSchemaMixin):
    """Manual preparation based on exported and imported metadata (and PDFs if any)"""

    settings: ExportManPrepSettings
    ci_supported: bool = False

    RELATIVE_PREP_MAN_PATH = Path("records_prep_man.bib")
    RELATIVE_PREP_MAN_INFO_PATH = Path("records_prep_man_info.csv")
    RELATIVE_PREP_MAN_INFO_PATH_XLS = Path("records_prep_man_info.xlsx")

    __FIELDS_TO_KEEP = [
        "ENTRYTYPE",
        "author",
        "title",
        "year",
        "journal",
        "booktitle",
        "incollection",
        "colrev_status",
        "volume",
        "number",
        "pages",
        "doi",
        "file",
    ]

    @dataclass
    class ExportManPrepSettings(
        colrev.env.package_manager.DefaultSettings, JsonSchemaMixin
    ):
        """Settings for ExportManPrep"""

        endpoint: str
        pdf_handling_mode: str = "symlink"

        _details = {
            "pdf_handling_mode": {
                "tooltip": "Indicates how linked PDFs are handled (symlink/copy_first_page)"
            },
        }

    settings_class = ExportManPrepSettings

    def __init__(
        self,
        *,
        prep_man_operation: colrev.ops.prep_man.PrepMan,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        if "pdf_handling_mode" not in settings:
            settings["pdf_handling_mode"] = "symlink"
        assert settings["pdf_handling_mode"] in ["symlink", "copy_first_page"]

        self.settings = self.settings_class.load_settings(data=settings)

        self.review_manager = prep_man_operation.review_manager
        self.quality_model = self.review_manager.get_qm()
        self.prep_man_bib_path = (
            self.review_manager.prep_dir / self.RELATIVE_PREP_MAN_PATH
        )

        self.prep_man_csv_path = (
            self.review_manager.prep_dir / self.RELATIVE_PREP_MAN_INFO_PATH
        )

        self.prep_man_xlsx_path = (
            self.review_manager.prep_dir / self.RELATIVE_PREP_MAN_INFO_PATH_XLS
        )

        self.review_manager.prep_dir.mkdir(exist_ok=True, parents=True)

    def __copy_files_for_man_prep(self, *, records: dict) -> None:
        prep_man_path_pdfs = self.review_manager.prep_dir / Path("pdfs")
        if prep_man_path_pdfs.is_dir():
            input(f"Remove {prep_man_path_pdfs} and press Enter.")
        prep_man_path_pdfs.mkdir(exist_ok=True, parents=True)

        for record in records.values():
            if "file" in record:
                target_path = self.review_manager.prep_dir / Path(record["file"])
                target_path.parents[0].mkdir(exist_ok=True, parents=True)

                if self.settings.pdf_handling_mode == "symlink":
                    try:
                        target_path.symlink_to(Path(record["file"]).resolve())
                    except FileExistsError:
                        pass

                if self.settings.pdf_handling_mode == "copy_first_page":
                    pdf_reader = PdfFileReader(str(record["file"]), strict=False)
                    if len(pdf_reader.pages) >= 1:
                        writer = PdfFileWriter()
                        writer.addPage(pdf_reader.pages[0])
                        with open(target_path, "wb") as outfile:
                            writer.write(outfile)

    def __export_prep_man(
        self,
        *,
        records: typing.Dict[str, typing.Dict],
    ) -> None:
        man_prep_recs = {
            k: v
            for k, v in records.items()
            if colrev.record.RecordState.md_needs_manual_preparation
            == v["colrev_status"]
        }

        # Filter out fields that are not needed for manual preparation
        filtered_man_prep_recs = {}
        for citation, fields in man_prep_recs.copy().items():
            for key in fields.copy():
                if key not in self.__FIELDS_TO_KEEP:
                    del fields[key]
            filtered_man_prep_recs.update({citation: fields})

        self.review_manager.dataset.save_records_dict_to_file(
            records=filtered_man_prep_recs, save_path=self.prep_man_bib_path
        )
        if any("file" in r for r in man_prep_recs.values()):
            self.__copy_files_for_man_prep(records=man_prep_recs)

    def __create_info_dataframe(
        self,
        *,
        records: typing.Dict[str, typing.Dict],
    ) -> None:
        man_prep_recs = [
            v
            for _, v in records.items()
            if colrev.record.RecordState.md_needs_manual_preparation
            == v["colrev_status"]
        ]

        man_prep_info = []
        for record in man_prep_recs:
            for field, value in record["colrev_masterdata_provenance"].items():
                if value["note"]:
                    man_prep_info.append(
                        {"ID": record["ID"], "field": field, "note": value["note"]}
                    )

        man_prep_info_df = pd.DataFrame(man_prep_info)
        if platform.system() == "Windows":
            # until https://github.com/pylint-dev/pylint/issues/3060 is resolved
            # pylint: disable=abstract-class-instantiated
            with pd.ExcelWriter(self.prep_man_xlsx_path) as writer:
                man_prep_info_df.to_excel(writer, index=False)
        else:
            man_prep_info_df.to_csv(self.prep_man_csv_path, index=False)

    def __drop_unnecessary_provenance_fiels(
        self, *, record: colrev.record.Record
    ) -> None:
        colrev_data_provenance_keys_to_drop = []
        for key, items in record.data.get("colrev_data_provenance", {}).items():
            if key not in record.data and "not-missing" not in items["note"]:
                colrev_data_provenance_keys_to_drop.append(key)
        for colrev_data_provenance_key_to_drop in colrev_data_provenance_keys_to_drop:
            del record.data["colrev_data_provenance"][
                colrev_data_provenance_key_to_drop
            ]

        colrev_masterdata_provenance_keys_to_drop = []
        for key, items in record.data.get("colrev_masterdata_provenance", {}).items():
            if key not in record.data and "not-missing" not in items["note"]:
                colrev_masterdata_provenance_keys_to_drop.append(key)
        for (
            colrev_masterdata_provenance_key_to_drop
        ) in colrev_masterdata_provenance_keys_to_drop:
            del record.data["colrev_masterdata_provenance"][
                colrev_masterdata_provenance_key_to_drop
            ]

    def __update_original_record_based_on_man_prepped(
        self, *, original_record: colrev.record.Record, man_prepped_record_dict: dict
    ) -> None:
        dropped_keys = [
            k
            for k in original_record.data
            if (k not in man_prepped_record_dict) and k in self.__FIELDS_TO_KEEP
        ]

        if original_record.data["ENTRYTYPE"] != man_prepped_record_dict["ENTRYTYPE"]:
            original_record.change_entrytype(
                new_entrytype=man_prepped_record_dict["ENTRYTYPE"],
                qm=self.quality_model,
            )

        if man_prepped_record_dict["ENTRYTYPE"] != original_record.data["ENTRYTYPE"]:
            original_record.data["ENTRYTYPE"] = man_prepped_record_dict["ENTRYTYPE"]
            original_record.update_masterdata_provenance(qm=self.quality_model)

        for key, value in man_prepped_record_dict.items():
            if key in ["colrev_status"]:
                continue
            if value != original_record.data.get(key, "") and value != "UNKNOWN":
                original_record.update_field(
                    key=key, value=value, source="man_prep", append_edit=False
                )
                original_record.remove_masterdata_provenance_note(
                    key=key, note="missing"
                )

        for dropped_key in dropped_keys:
            original_record.remove_field(
                key=dropped_key, not_missing_note=True, source="man_prep"
            )

    def __print_stats(self, *, original_record: colrev.record.Record) -> None:
        if (
            original_record.data["colrev_status"]
            == colrev.record.RecordState.rev_prescreen_excluded
        ):
            self.review_manager.logger.info(
                f" {colors.RED}{original_record.data['ID']}".ljust(46)
                + "md_needs_manual_preparation →  rev_prescreen_excluded"
                + f"{colors.END}"
            )
        elif (
            original_record.data["colrev_status"]
            == colrev.record.RecordState.md_prepared
        ):
            self.review_manager.logger.info(
                f" {colors.GREEN}{original_record.data['ID']}".ljust(46)
                + "md_needs_manual_preparation →  md_prepared"
                + f"{colors.END}"
            )
        else:
            man_prep_note = ", ".join(
                k + ":" + v["note"]
                for k, v in original_record.data["colrev_masterdata_provenance"].items()
                if v["note"] != ""
            )
            self.review_manager.logger.info(
                f" {colors.ORANGE}{original_record.data['ID']}".ljust(46)
                + f"{man_prep_note}"
                + f"{colors.END}"
            )

    def __import_record(
        self,
        *,
        man_prepped_record_dict: dict,
        original_record: colrev.record.Record,
        imported_records: list,
    ) -> None:
        imported_records.append(original_record.data["ID"])
        override = (
            man_prepped_record_dict["colrev_status"]
            == colrev.record.RecordState.md_prepared
        )

        self.__update_original_record_based_on_man_prepped(
            original_record=original_record,
            man_prepped_record_dict=man_prepped_record_dict,
        )

        self.__drop_unnecessary_provenance_fiels(record=original_record)
        if (
            man_prepped_record_dict["colrev_status"]
            == colrev.record.RecordState.rev_prescreen_excluded
        ):
            original_record.set_status(
                target_state=colrev.record.RecordState.rev_prescreen_excluded,
                force=True,
            )

        else:
            original_record.update_masterdata_provenance(
                qm=self.quality_model, set_prepared=True
            )

        if override:
            original_record.set_status(
                target_state=colrev.record.RecordState.md_prepared, force=True
            )

        self.__print_stats(original_record=original_record)

    def __import_prep_man(self) -> None:
        self.review_manager.logger.info(
            "Load import changes from "
            f"{self.prep_man_bib_path.relative_to(self.review_manager.path)}"
        )

        with open(self.prep_man_bib_path, encoding="utf8") as target_bib:
            man_prep_recs = self.review_manager.dataset.load_records_dict(
                load_str=target_bib.read()
            )

        imported_records: typing.List[dict] = []
        records = self.review_manager.dataset.load_records_dict()
        for record_id, record_dict in records.items():
            if (
                record_dict["colrev_status"]
                == colrev.record.RecordState.rev_prescreen_excluded
                # or record_id not in man_prep_recs
            ):
                records[record_id][  # pylint: disable=direct-status-assign
                    "colrev_status"
                ] = colrev.record.RecordState.rev_prescreen_excluded
                self.review_manager.logger.info(
                    f" {colors.RED}{record_id}".ljust(46)
                    + "md_needs_manual_preparation →  rev_prescreen_excluded"
                    + f"{colors.END}"
                )

        for record_id, man_prepped_record_dict in man_prep_recs.items():
            if record_id not in records:
                print(f"ID no longer in records: {record_id}")
                continue
            original_record = colrev.record.Record(
                data=records[man_prepped_record_dict["ID"]]
            )
            self.__import_record(
                man_prepped_record_dict=man_prepped_record_dict,
                original_record=original_record,
                imported_records=imported_records,
            )

        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()
        self.review_manager.create_commit(msg="Prep-man (ExportManPrep)")

        self.review_manager.dataset.set_ids(selected_ids=imported_records)
        self.review_manager.create_commit(msg="Set IDs")

    def __print_export_prep_man_instructions(self) -> None:
        print("Created two files:")
        if platform.system() == "Windows":
            print(
                f" - {self.prep_man_xlsx_path.relative_to(self.review_manager.path)}  (EXCEL file)"
            )
        else:
            print(
                f" - {self.prep_man_csv_path.relative_to(self.review_manager.path)}  (CSV file)"
            )
        print(
            f" - {self.prep_man_bib_path.relative_to(self.review_manager.path)}       (BIB file)"
        )
        print()
        print("To prepare records:")
        print("- check the defect codes in the CSV file")
        print(
            "- edit the BIB file and change the fields "
            "(e.g., add missing volume/number or remove all-caps)"
        )
        print(
            "- if a record should be exluded in the prescreen, simply remove it from the BIB file"
        )
        print()

        print(f"Once completed, run {colors.ORANGE}colrev prep-man{colors.END} again.")

    def prepare_manual(
        self, prep_man_operation: colrev.ops.prep_man.PrepMan, records: dict
    ) -> dict:
        """Prepare records manually by extracting the subset of records to a separate BiBTex file"""

        if not self.prep_man_bib_path.is_file():
            self.__create_info_dataframe(records=records)
            self.__export_prep_man(records=records)
            self.__print_export_prep_man_instructions()
        else:
            selected_path = self.prep_man_bib_path.relative_to(
                prep_man_operation.review_manager.path
            )
            if input(f"Import changes from {selected_path} [y,n]?") == "y":
                self.__import_prep_man()

        return records



#! /usr/bin/env python
"""SearchSource: ACM Digital Library"""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.load_utils_bib
import colrev.ops.search
import colrev.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class ACMDigitalLibrarySearchSource(JsonSchemaMixin):
    """SearchSource for the ACM digital Library"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    # Note : the ID contains the doi
    # "https://dl.acm.org/doi/{{ID}}"
    source_identifier = "doi"
    search_type = colrev.settings.SearchType.DB
    api_search_supported = False
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "ACM Digital Library"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/colrev/"
        + "ops/built_in/search_sources/acm_digital_library.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.Operation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for ACM dDigital Library"""

        result = {"confidence": 0.0}
        # Simple heuristic:
        if "publisher = {Association for Computing Machinery}," in data:
            result["confidence"] = 0.7
            print(data)
            return result
        # We may also check whether the ID=doi=url
        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""
        raise NotImplementedError

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of ACM Digital Library"""

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.ops.load_utils_bib.load_bib_file(
                load_operation=load_operation, source=self.search_source
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for ACM Digital Library"""
        record.remove_field(key="url")
        record.remove_field(key="numpages")
        record.remove_field(key="issue_date")
        record.remove_field(key="publisher")
        record.remove_field(key="address")
        record.remove_field(key="month")

        return record



#! /usr/bin/env python
"""SearchSource: IEEEXplore"""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.built_in.search_sources.ieee_api
import colrev.ops.load_utils_ris
import colrev.ops.prep
import colrev.ops.search
import colrev.record


# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class IEEEXploreSearchSource(JsonSchemaMixin):
    """SearchSource for IEEEXplore"""

    flag = True
    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "ID"
    search_type = colrev.settings.SearchType.DB
    endpoint = "colrev.ieee"
    api_search_supported = True
    ci_supported: bool = True
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.oni
    short_name = "IEEE Xplore"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/ieee.md"
    )
    SETTINGS = {
        "api_key": "packages.search_source.colrev.ieee.api_key",
    }

    API_FIELDS = [
        "abstract",
        "author_url",
        "accessType",
        "article_number",
        "author_order",
        "author_terms",
        "affiliation",
        "citing_paper_count",
        "conference_dates",
        "conference_location",
        "content_type",
        "doi",
        "publisher",
        "pubtype",
        "d-year",
        "end_page",
        "facet",
        "full_name",
        "html_url",
        "ieee_terms",
        "isbn",
        "issn",
        "issue",
        "pdf_url",
        "publication_year",
        "publication_title",
        "standard_number",
        "standard_status",
        "start_page",
        "title",
        "totalfound",
        "totalsearched",
        "volume",
    ]

    FIELD_MAPPING = {
        "citing_paper_count": "citations",
        "publication_year": "year",
        "html_url": "url",
        "pdf_url": "fulltext",
        "issue": "number",
    }

    def __init__(
        self,
        *,
        source_operation: colrev.operation.Operation,
        settings: Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager

        if settings:
            self.search_source = from_dict(
                data_class=self.settings_class, data=settings
            )
        else:
            self.search_source = colrev.settings.SearchSource(
                endpoint=self.endpoint,
                filename=Path("data/search/ieee.bib"),
                search_type=colrev.settings.SearchType.OTHER,
                search_parameters={},
                comment="",
            )

    # For run_search, a Python SDK would be available:
    # https://developer.ieee.org/Python_Software_Development_Kit

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for IEEEXplore"""

        result = {"confidence": 0.1}

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""

        if params is None:
            operation.add_interactively(endpoint=cls.endpoint)
            return

        if "https://ieeexploreapi.ieee.org/api/v1/search/articles?" in params:
            params = params.replace(
                "https://ieeexploreapi.ieee.org/api/v1/search/articles?", ""
            ).lstrip("&")

            parameter_pairs = params.split("&")
            search_parameters = {}
            for parameter in parameter_pairs:
                key, value = parameter.split("=")
                search_parameters[key] = value

            last_value = list(search_parameters.values())[-1]

            filename = operation.get_unique_filename(
                file_path_string=f"ieee_{last_value}"
            )

            add_source = colrev.settings.SearchSource(
                endpoint=cls.endpoint,
                filename=filename,
                search_type=colrev.settings.SearchType.DB,
                search_parameters=search_parameters,
                comment="",
            )
            operation.review_manager.settings.sources.append(add_source)
            return

        raise NotImplementedError

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of IEEEXplore"""

        # pylint: disable=too-many-locals

        ieee_feed = self.search_source.get_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        api_key = self.review_manager.environment_manager.get_settings_by_key(
            self.SETTINGS["api_key"]
        )
        if api_key is None or len(api_key) != 24:
            api_key = input("Please enter api key: ")
            self.review_manager.environment_manager.update_registry(
                self.SETTINGS["api_key"], api_key
            )

        query = colrev.ops.built_in.search_sources.ieee_api.XPLORE(api_key)
        query.dataType("json")
        query.dataFormat("object")
        query.maximumResults(50000)
        query.usingOpenAccess = False

        parameter_methods = {}
        parameter_methods["article_number"] = query.articleNumber
        parameter_methods["doi"] = query.doi
        parameter_methods["author"] = query.authorText
        parameter_methods["isbn"] = query.isbn
        parameter_methods["issn"] = query.issn
        parameter_methods["publication_year"] = query.publicationYear
        parameter_methods["queryText"] = query.queryText
        parameter_methods["parameter"] = query.queryText

        parameters = self.search_source.search_parameters
        for key, value in parameters.items():
            if key in parameter_methods:
                method = parameter_methods[key]
                method(value)

        query.startRecord = 1
        response = query.callAPI()
        while "articles" in response:
            records = self.review_manager.dataset.load_records_dict()
            articles = response["articles"]

            for article in articles:
                prev_record_dict_version: dict = {}
                record_dict = self.__create_record_dict(article)
                record = colrev.record.Record(data=record_dict)
                added = ieee_feed.add_record(record=record)

                if added:
                    self.review_manager.logger.info(" retrieve " + record.data["ID"])
                else:
                    changed = ieee_feed.update_existing_record(
                        records=records,
                        record_dict=record.data,
                        prev_record_dict_version=prev_record_dict_version,
                        source=self.search_source,
                        update_time_variant_fields=rerun,
                    )
                    if changed:
                        self.review_manager.logger.info(" update " + record.data["ID"])

            query.startRecord += 200
            response = query.callAPI()

        ieee_feed.print_post_run_search_infos(records=records)
        ieee_feed.save_feed_file()
        self.review_manager.dataset.save_records_dict(records=records)
        self.review_manager.dataset.add_record_changes()

    def __update_special_case_fields(self, *, record_dict: dict, article: dict) -> None:
        if "start_page" in article:
            record_dict["pages"] = article.pop("start_page")
            if "end_page" in article:
                record_dict["pages"] += "--" + article.pop("end_page")

        if "authors" in article and "authors" in article["authors"]:
            author_list = []
            for author in article["authors"]["authors"]:
                author_list.append(author["full_name"])
            record_dict["author"] = colrev.record.PrepRecord.format_author_field(
                input_string=" and ".join(author_list)
            )

        if (
            "index_terms" in article
            and "author_terms" in article["index_terms"]
            and "terms" in article["index_terms"]["author_terms"]
        ):
            record_dict["keywords"] = ", ".join(
                article["index_terms"]["author_terms"]["terms"]
            )

    def __create_record_dict(self, article: dict) -> dict:
        record_dict = {"ID": article["article_number"]}
        # self.review_manager.p_printer.pprint(article)

        if article["content_type"] == "Conferences":
            record_dict["ENTRYTYPE"] = "inproceedings"
            if "publication_title" in article:
                record_dict["booktitle"] = article.pop("publication_title")
        else:
            record_dict["ENTRYTYPE"] = "article"
            if "publication_title" in article:
                record_dict["journal"] = article.pop("publication_title")

        for field in self.API_FIELDS:
            if article.get(field) is None:
                continue
            record_dict[field] = str(article.get(field))

        for api_field, rec_field in self.FIELD_MAPPING.items():
            if api_field not in record_dict:
                continue
            record_dict[rec_field] = record_dict.pop(api_field)

        self.__update_special_case_fields(record_dict=record_dict, article=article)

        return record_dict

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def __ris_fixes(self, *, entries: dict) -> None:
        for entry in entries:
            if entry["type_of_reference"] in ["CONF", "JOUR"]:
                if "title" in entry and "primary_title" not in entry:
                    entry["primary_title"] = entry.pop("title")
            if "publication_year" in entry and "year" not in entry:
                entry["year"] = entry.pop("publication_year")

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".ris":
            ris_loader = colrev.ops.load_utils_ris.RISLoader(
                load_operation=load_operation, source=self.search_source
            )
            ris_entries = ris_loader.load_ris_entries(
                filename=self.search_source.filename
            )
            self.__ris_fixes(entries=ris_entries)
            records = ris_loader.convert_to_records(entries=ris_entries)
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for IEEEXplore"""

        return record



#! /usr/bin/env python
"""Preparation of curations"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.search_sources
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class CurationPrep(JsonSchemaMixin):
    """Preparation of curations"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)
        self.quality_model = prep_operation.review_manager.get_qm()
        self.prep_operation = prep_operation

    def prepare(
        self,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        record: colrev.record.PrepRecord,
    ) -> colrev.record.Record:
        """Prepare records in a CoLRev curation"""

        # pylint: disable=too-many-branches

        if (
            record.data["colrev_status"]
            == colrev.record.RecordState.rev_prescreen_excluded
        ):
            return record

        if record.data.get("year", "UNKNOWN") == "UNKNOWN":
            record.set_status(
                target_state=colrev.record.RecordState.md_needs_manual_preparation
            )
            colrev.record.Record(data=record.data).add_masterdata_provenance(
                key="year",
                source="colrev_curation.masterdata_restrictions",
                note="missing",
            )
            return record

        colrev.record.Record(data=record.data).update_masterdata_provenance(
            qm=self.quality_model
        )

        return record



#! /usr/bin/env python
"""Removal of broken URLs (error 500) a prep operation"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import requests
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.search_sources
import colrev.record

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class RemoveError500URLsPrep(JsonSchemaMixin):
    """Prepares records by removing urls that are not available"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

    def prepare(
        self, prep_operation: colrev.ops.prep.Prep, record: colrev.record.PrepRecord
    ) -> colrev.record.Record:
        """Prepare the record by removing URLs with 500 errors"""

        session = prep_operation.review_manager.get_cached_session()

        try:
            if "url" in record.data:
                ret = session.request(
                    "GET",
                    record.data["url"],
                    headers=prep_operation.requests_headers,
                    timeout=60,
                )
                if ret.status_code >= 500:
                    record.remove_field(key="url")
        except requests.exceptions.RequestException:
            pass
        try:
            if "fulltext" in record.data:
                ret = session.request(
                    "GET",
                    record.data["fulltext"],
                    headers=prep_operation.requests_headers,
                    timeout=prep_operation.timeout,
                )
                if ret.status_code >= 500:
                    record.remove_field(key="fulltext")
        except requests.exceptions.RequestException:
            pass

        return record



#! /usr/bin/env python
"""SearchSource: Unknown source (default for all other sources)"""
from __future__ import annotations

import re
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

import dacite
import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin
from thefuzz import fuzz

import colrev.env.language_service
import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.load_utils_bib
import colrev.ops.load_utils_md
import colrev.ops.load_utils_ris
import colrev.ops.search
import colrev.record
import colrev.ui_cli.cli_colors as colors

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class UnknownSearchSource(JsonSchemaMixin):
    """SearchSource for unknown search results"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings

    source_identifier = "colrev.unknown_source"
    search_type = colrev.settings.SearchType.DB
    api_search_supported = False
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.na
    short_name = "Unknown Source"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/unknown_source.md"
    )

    HTML_CLEANER = re.compile("<.*?>")
    __padding = 40

    def __init__(
        self, *, source_operation: colrev.operation.Operation, settings: dict
    ) -> None:
        converters = {Path: Path, Enum: Enum}
        self.search_source = from_dict(
            data_class=self.settings_class,
            data=settings,
            config=dacite.Config(type_hooks=converters, cast=[Enum]),  # type: ignore
        )
        self.review_manager = source_operation.review_manager
        self.language_service = colrev.env.language_service.LanguageService()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for unknown sources"""

        result = {"confidence": 0.1}

        return result

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.search.Search, params: str) -> None:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""
        raise NotImplementedError

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of Crossref"""

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def __ris_fixes(self, *, entries: dict) -> None:
        for entry in entries:
            if "title" in entry and "primary_title" not in entry:
                entry["primary_title"] = entry.pop("title")

            if "publication_year" in entry and "year" not in entry:
                entry["year"] = entry.pop("publication_year")

    def __rename_erroneous_extensions(self) -> None:
        if self.search_source.filename.suffix in [".xls", ".xlsx"]:
            return
        data = self.search_source.filename.read_text(encoding="utf-8")
        # # Correct the file extension if necessary
        if re.findall(
            r"^%0", data, re.MULTILINE
        ) and self.search_source.filename.suffix not in [".enl"]:
            new_filename = self.search_source.filename.with_suffix(".enl")
            self.review_manager.logger.info(
                f"{colors.GREEN}Rename to {new_filename} "
                f"(because the format is .enl){colors.END}"
            )
            self.search_source.filename.rename(new_filename)
            self.review_manager.dataset.add_changes(
                path=self.search_source.filename, remove=True
            )
            self.search_source.filename = new_filename
            self.review_manager.dataset.add_changes(path=new_filename)
            self.review_manager.create_commit(
                msg=f"Rename {self.search_source.filename}"
            )
            return

        if re.findall(
            r"^TI ", data, re.MULTILINE
        ) and self.search_source.filename.suffix not in [".ris"]:
            new_filename = self.search_source.filename.with_suffix(".ris")
            self.review_manager.logger.info(
                f"{colors.GREEN}Rename to {new_filename} "
                f"(because the format is .ris){colors.END}"
            )
            self.search_source.filename.rename(new_filename)
            self.review_manager.dataset.add_changes(
                path=self.search_source.filename, remove=True
            )
            self.search_source.filename = new_filename
            self.review_manager.dataset.add_changes(path=new_filename)
            self.review_manager.create_commit(
                msg=f"Rename {self.search_source.filename}"
            )

    def __load_ris(self, *, load_operation: colrev.ops.load.Load) -> dict:
        ris_loader = colrev.ops.load_utils_ris.RISLoader(
            load_operation=load_operation, source=self.search_source
        )
        ris_loader.apply_ris_fixes(filename=self.search_source.filename)
        ris_entries = ris_loader.load_ris_entries(filename=self.search_source.filename)
        self.__ris_fixes(entries=ris_entries)
        records = ris_loader.convert_to_records(entries=ris_entries)
        return records

    def __load_bib(self, *, load_operation: colrev.ops.load.Load) -> dict:
        records = colrev.ops.load_utils_bib.load_bib_file(
            load_operation=load_operation, source=self.search_source
        )
        return records

    def __load_csv(self, *, load_operation: colrev.ops.load.Load) -> dict:
        csv_loader = colrev.ops.load_utils_table.CSVLoader(
            load_operation=load_operation, source=self.search_source
        )
        records = csv_loader.load()
        return records

    def __load_xlsx(self, *, load_operation: colrev.ops.load.Load) -> dict:
        excel_loader = colrev.ops.load_utils_table.ExcelLoader(
            load_operation=load_operation, source=self.search_source
        )
        records = excel_loader.load()
        return records

    def __load_md(self, *, load_operation: colrev.ops.load.Load) -> dict:
        md_loader = colrev.ops.load_utils_md.MarkdownLoader(
            load_operation=load_operation, source=self.search_source
        )
        records = md_loader.load()
        return records

    def __load_enl(self, *, load_operation: colrev.ops.load.Load) -> dict:
        enl_loader = colrev.ops.load_utils_enl.ENLLoader(
            load_operation=load_operation, source=self.search_source
        )
        records = enl_loader.load(source=self.search_source)
        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if not self.search_source.filename.is_file():
            return {}

        self.__rename_erroneous_extensions()

        __load_methods = {
            ".ris": self.__load_ris,
            ".bib": self.__load_bib,
            ".csv": self.__load_csv,
            ".xls": self.__load_xlsx,
            ".xlsx": self.__load_xlsx,
            ".md": self.__load_md,
            ".enl": self.__load_enl,
        }

        if self.search_source.filename.suffix not in __load_methods:
            raise NotImplementedError

        return __load_methods[self.search_source.filename.suffix](
            load_operation=load_operation
        )

    def __heuristically_fix_entrytypes(
        self, *, record: colrev.record.PrepRecord
    ) -> None:
        """Prepare the record by heuristically correcting erroneous ENTRYTYPEs"""

        # Journal articles should not have booktitles/series set.
        if record.data["ENTRYTYPE"] == "article":
            if "booktitle" in record.data and "journal" not in record.data:
                record.update_field(
                    key="journal",
                    value=record.data["booktitle"],
                    source="unkown_source_prep",
                )
                record.remove_field(key="booktitle")
            if "series" in record.data and "journal" not in record.data:
                record.update_field(
                    key="journal",
                    value=record.data["series"],
                    source="unkown_source_prep",
                )
                record.remove_field(key="series")

        if self.search_source.filename.suffix == ".md":
            if record.data["ENTRYTYPE"] == "misc" and "publisher" in record.data:
                record.update_field(
                    key="ENTRYTYPE", value="book", source="unkown_source_prep"
                )
            if record.data.get("year", "year") == record.data.get("date", "date"):
                record.remove_field(key="date")
            if (
                "inbook" == record.data["ENTRYTYPE"]
                and "chapter" not in record.data
                and "title" in record.data
            ):
                record.rename_field(key="title", new_key="chapter")

        if (
            "dissertation" in record.data.get("fulltext", "NA").lower()
            and record.data["ENTRYTYPE"] != "phdthesis"
        ):
            prior_e_type = record.data["ENTRYTYPE"]
            record.update_field(
                key="ENTRYTYPE", value="phdthesis", source="unkown_source_prep"
            )
            self.review_manager.report_logger.info(
                f' {record.data["ID"]}'.ljust(self.__padding, " ")
                + f"Set from {prior_e_type} to phdthesis "
                '("dissertation" in fulltext link)'
            )

        if (
            "thesis" in record.data.get("fulltext", "NA").lower()
            and record.data["ENTRYTYPE"] != "phdthesis"
        ):
            prior_e_type = record.data["ENTRYTYPE"]
            record.update_field(
                key="ENTRYTYPE", value="phdthesis", source="unkown_source_prep"
            )
            self.review_manager.report_logger.info(
                f' {record.data["ID"]}'.ljust(self.__padding, " ")
                + f"Set from {prior_e_type} to phdthesis "
                '("thesis" in fulltext link)'
            )

        if (
            "this thesis" in record.data.get("abstract", "NA").lower()
            and record.data["ENTRYTYPE"] != "phdthesis"
        ):
            prior_e_type = record.data["ENTRYTYPE"]
            record.update_field(
                key="ENTRYTYPE", value="phdthesis", source="unkown_source_prep"
            )
            self.review_manager.report_logger.info(
                f' {record.data["ID"]}'.ljust(self.__padding, " ")
                + f"Set from {prior_e_type} to phdthesis "
                '("thesis" in abstract)'
            )

    def __format_inproceedings(self, *, record: colrev.record.PrepRecord) -> None:
        if record.data.get("booktitle", "UNKNOWN") == "UNKNOWN":
            return

        if (
            "UNKNOWN" != record.data["booktitle"]
            and "inbook" != record.data["ENTRYTYPE"]
        ):
            record.format_if_mostly_upper(key="booktitle", case="title")

            stripped_btitle = re.sub(r"\d{4}", "", record.data["booktitle"])
            stripped_btitle = re.sub(r"\d{1,2}th", "", stripped_btitle)
            stripped_btitle = re.sub(r"\d{1,2}nd", "", stripped_btitle)
            stripped_btitle = re.sub(r"\d{1,2}rd", "", stripped_btitle)
            stripped_btitle = re.sub(r"\d{1,2}st", "", stripped_btitle)
            stripped_btitle = re.sub(r"\([A-Z]{3,6}\)", "", stripped_btitle)
            stripped_btitle = stripped_btitle.replace("Proceedings of the", "").replace(
                "Proceedings", ""
            )
            stripped_btitle = stripped_btitle.lstrip().rstrip()
            record.update_field(
                key="booktitle",
                value=stripped_btitle,
                source="unkown_source_prep",
                keep_source_if_equal=True,
            )

    def __format_article(self, record: colrev.record.PrepRecord) -> None:
        if (
            record.data.get("journal", "UNKNOWN") != "UNKNOWN"
            and len(record.data["journal"]) > 10
            and "UNKNOWN" != record.data["journal"]
        ):
            record.format_if_mostly_upper(key="journal", case="title")

        if record.data.get("volume", "UNKNOWN") != "UNKNOWN":
            record.update_field(
                key="volume",
                value=record.data["volume"].replace("Volume ", ""),
                source="unkown_source_prep",
                keep_source_if_equal=True,
            )

    def __format_fields(self, *, record: colrev.record.PrepRecord) -> None:
        """Format fields"""

        if record.data.get("ENTRYTYPE", "") == "inproceedings":
            self.__format_inproceedings(record=record)
        elif record.data.get("ENTRYTYPE", "") == "article":
            self.__format_article(record=record)

        if record.data.get("author", "UNKNOWN") != "UNKNOWN":
            # fix name format
            if (1 == len(record.data["author"].split(" ")[0])) or (
                ", " not in record.data["author"]
            ):
                record.update_field(
                    key="author",
                    value=colrev.record.PrepRecord.format_author_field(
                        input_string=record.data["author"]
                    ),
                    source="unkown_source_prep",
                    keep_source_if_equal=True,
                )
            # Replace nicknames in parentheses
            record.data["author"] = re.sub(r"\([^)]*\)", "", record.data["author"])
            record.data["author"] = record.data["author"].replace("  ", " ").rstrip()

        if record.data.get("title", "UNKNOWN") != "UNKNOWN":
            record.format_if_mostly_upper(key="title")

        if "pages" in record.data:
            record.unify_pages_field()
            if (
                not re.match(r"^\d*$", record.data["pages"])
                and not re.match(r"^\d*--\d*$", record.data["pages"])
                and not re.match(r"^[xivXIV]*--[xivXIV]*$", record.data["pages"])
            ):
                self.review_manager.report_logger.info(
                    f' {record.data["ID"]}:'.ljust(self.__padding, " ")
                    + f'Unusual pages: {record.data["pages"]}'
                )

        if (
            "url" in record.data
            and "fulltext" in record.data
            and record.data["url"] == record.data["fulltext"]
        ):
            record.remove_field(key="fulltext")

        if "language" in record.data:
            try:
                self.language_service.unify_to_iso_639_3_language_codes(record=record)
                record.update_field(
                    key="language",
                    value=record.data["language"],
                    source="unkown_source_prep",
                    keep_source_if_equal=True,
                )
            except colrev_exceptions.InvalidLanguageCodeException:
                del record.data["language"]

    def __remove_redundant_fields(self, *, record: colrev.record.PrepRecord) -> None:
        if (
            record.data["ENTRYTYPE"] == "article"
            and "journal" in record.data
            and "booktitle" in record.data
        ):
            similarity_journal_booktitle = fuzz.partial_ratio(
                record.data["journal"].lower(), record.data["booktitle"].lower()
            )
            if similarity_journal_booktitle / 100 > 0.9:
                record.remove_field(key="booktitle")

        if record.data.get("publisher", "") in ["researchgate.net"]:
            record.remove_field(key="publisher")

        if (
            record.data["ENTRYTYPE"] == "inproceedings"
            and "journal" in record.data
            and "booktitle" in record.data
        ):
            similarity_journal_booktitle = fuzz.partial_ratio(
                record.data["journal"].lower(), record.data["booktitle"].lower()
            )
            if similarity_journal_booktitle / 100 > 0.9:
                record.remove_field(key="journal")

    def __impute_missing_fields(self, *, record: colrev.record.PrepRecord) -> None:
        if "date" in record.data and "year" not in record.data:
            year = re.search(r"\d{4}", record.data["date"])
            if year:
                record.update_field(
                    key="year",
                    value=year.group(0),
                    source="unkown_source_prep",
                    keep_source_if_equal=True,
                )

    def __unify_special_characters(self, *, record: colrev.record.PrepRecord) -> None:
        # Remove html entities
        for field in list(record.data.keys()):
            # Skip dois (and their provenance), which may contain html entities
            if field in [
                "colrev_masterdata_provenance",
                "colrev_data_provenance",
                "doi",
            ]:
                continue
            if field in ["author", "title", "journal"]:
                record.data[field] = re.sub(r"\s+", " ", record.data[field])
                record.data[field] = re.sub(self.HTML_CLEANER, "", record.data[field])

    def prepare(
        self, record: colrev.record.PrepRecord, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for unknown sources"""

        if not record.has_quality_defects() or record.masterdata_is_curated():
            return record

        # TODO : assign fields (e.g., to colrev.pubmed.pubmedid)

        self.__heuristically_fix_entrytypes(
            record=record,
        )

        self.__impute_missing_fields(record=record)

        self.__format_fields(record=record)

        self.__remove_redundant_fields(record=record)

        self.__unify_special_characters(record=record)

        return record



#! /usr/bin/env python
"""Exclude records with non-latin alphabets as a prep operation"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import zope.interface
from alphabet_detector import AlphabetDetector
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.search_sources
import colrev.record

# pylint: disable=duplicate-code

if TYPE_CHECKING:
    import colrev.ops.prep

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class ExcludeNonLatinAlphabetsPrep(JsonSchemaMixin):
    """Prepares records by excluding ones that have a non-latin alphabet
    (in the title, author, journal, or booktitle field)"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = "check with the developer"
    always_apply_changes = True
    alphabet_detector = AlphabetDetector()

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

    def prepare(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.PrepRecord,
    ) -> colrev.record.Record:
        """Prepare the records by excluding records whose metadata is not in Latin alphabet"""

        if prep_operation.polish:
            return record

        def mostly_latin_alphabet(str_to_check: str) -> bool:
            assert len(str_to_check) != 0
            nr_non_latin = 0
            for character in str_to_check:
                if not self.alphabet_detector.only_alphabet_chars(character, "LATIN"):
                    nr_non_latin += 1
            return nr_non_latin / len(str_to_check) > 0.75

        # TB:D join or check independently?
        str_to_check = " ".join(
            [
                record.data.get("title", ""),
                record.data.get("author", ""),
                record.data.get("journal", ""),
                record.data.get("booktitle", ""),
            ]
        )
        if mostly_latin_alphabet(str_to_check):
            record.prescreen_exclude(reason="non_latin_alphabet")

        return record



#! /usr/bin/env python
"""CoLRev init operation: Create a project and specify settings."""
from __future__ import annotations

import json
import logging
import os
import platform
from importlib.metadata import version
from pathlib import Path
from subprocess import CalledProcessError  # nosec
from subprocess import check_call  # nosec
from subprocess import DEVNULL  # nosec
from subprocess import STDOUT  # nosec
from typing import Optional

import git

import colrev.dataset
import colrev.env.environment_manager
import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.review_manager  # pylint: disable=cyclic-import
import colrev.settings
import colrev.ui_cli.cli_colors as colors


# pylint: disable=too-few-public-methods


class Initializer:
    """Initialize a CoLRev project"""

    share_stat_req_options = ["none", "processed", "screened", "completed"]

    def __init__(
        self,
        *,
        review_type: str,
        example: bool = False,
        light: bool = False,
        local_pdf_collection: bool = False,
        target_path: Optional[Path] = None,
        exact_call: str = "",
    ) -> None:
        saved_args = locals()
        if example and local_pdf_collection:
            raise colrev_exceptions.RepoInitError(
                msg="Cannot initialize local_pdf_collection repository with example data."
            )
        current_platform = platform.system()
        if current_platform != "Linux":
            light = True
        self.light = light

        self.review_type = review_type.replace("-", "_").lower().replace(" ", "_")
        if "." not in self.review_type:
            self.review_type = "colrev." + self.review_type

        if target_path:
            os.chdir(target_path)
            self.target_path = target_path
        else:
            self.target_path = Path.cwd()

        review_manager = colrev.review_manager.ReviewManager(
            path_str=str(target_path), force_mode=True, navigate_to_home_dir=False
        )

        try:
            res = review_manager.get_review_types(review_type=self.review_type)
        except colrev.exceptions.MissingDependencyError as exc:
            res = review_manager.get_review_types(
                review_type="colrev.literature_review"
            )
            raise colrev_exceptions.ParameterError(
                parameter="init.review_type",
                value=f"'{review_type}'",
                options=list(res.all_available_packages_names.keys()),
            ) from exc

        self.__check_init_precondition()

        self.title = str(self.target_path.name)
        self.logger = self.__setup_init_logger(level=logging.INFO)

        self.__setup_git()
        self.__setup_files(path=self.target_path)

        if example:
            self.__create_example_repo()

        self.review_manager = colrev.review_manager.ReviewManager(exact_call=exact_call)

        self.__create_commit(saved_args=saved_args)
        if not example:
            if "pytest" not in os.getcwd():
                self.__register_repo()
        if local_pdf_collection:
            self.__create_local_pdf_collection()

        self.__post_commit_edits()

        self.review_manager.logger.info(
            "%sCompleted init operation%s", colors.GREEN, colors.END
        )

    def __check_init_precondition(self) -> None:
        cur_content = [
            str(x.relative_to(self.target_path)) for x in self.target_path.glob("**/*")
        ]
        cur_content = [
            x for x in cur_content if not x.startswith("venv") and x != ".history"
        ]

        if str(colrev.review_manager.ReviewManager.REPORT_RELATIVE) in cur_content:
            cur_content.remove(str(colrev.review_manager.ReviewManager.REPORT_RELATIVE))
        if cur_content:
            raise colrev_exceptions.NonEmptyDirectoryError(
                filepath=self.target_path, content=cur_content
            )

        environment_manager = colrev.env.environment_manager.EnvironmentManager()
        environment_manager.get_name_mail_from_git()

        try:
            environment_manager.check_docker_installed()
        except colrev_exceptions.MissingDependencyError as exc:  # pragma: no cover
            if not self.light:
                raise colrev_exceptions.CoLRevException(
                    "Docker not installed. Docker is optional but recommended.\n"
                    "For more information, see "
                    "https://colrev.readthedocs.io/en/latest/manual/manual.html"
                    "To init a repository without Docker, run "
                    f"{colors.ORANGE}colrev init --light{colors.END}"
                ) from exc

    def __setup_init_logger(self, *, level: int = logging.INFO) -> logging.Logger:
        # pylint: disable=duplicate-code
        init_logger = logging.getLogger("colrev-init_logger")

        init_logger.setLevel(level)

        if init_logger.handlers:
            for handler in init_logger.handlers:
                init_logger.removeHandler(handler)

        formatter = logging.Formatter(
            fmt="%(asctime)s [%(levelname)s] %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        handler.setLevel(level)

        init_logger.addHandler(handler)
        init_logger.propagate = False

        return init_logger

    def __register_repo(self) -> None:
        self.review_manager.logger.info("Register CoLRev repository")

        environment_manager = self.review_manager.get_environment_manager()
        environment_manager.register_repo(path_to_register=self.target_path)

    def __create_commit(self, *, saved_args: dict) -> None:
        del saved_args["local_pdf_collection"]
        self.review_manager.create_commit(
            msg="Initial commit",
            manual_author=True,
        )

    def __setup_files(self, *, path: Path) -> None:
        # pylint: disable=too-many-locals

        # Note: parse instead of copy to avoid format changes
        settings_filedata = colrev.env.utils.get_package_file_content(
            file_path=Path("template/init/settings.json")
        )
        if settings_filedata:
            settings = json.loads(settings_filedata.decode("utf-8"))
            settings["project"]["review_type"] = str(self.review_type)
            with open(path / Path("settings.json"), "w", encoding="utf8") as file:
                json.dump(settings, file, indent=4)

        colrev.review_manager.ReviewManager.SEARCHDIR_RELATIVE.mkdir(parents=True)
        colrev.review_manager.ReviewManager.PDF_DIR_RELATIVE.mkdir(parents=True)

        colrev_path = Path.home() / Path("colrev")
        colrev_path.mkdir(exist_ok=True, parents=True)

        files_to_retrieve = [
            [Path("template/init/readme.md"), Path("readme.md")],
            [
                Path("template/init/pre-commit-config.yaml"),
                Path(".pre-commit-config.yaml"),
            ],
            [Path("template/init/markdownlint.yaml"), Path(".markdownlint.yaml")],
            [
                Path("template/init/pre-commit.yml"),
                Path(".github/workflows/pre-commit.yml"),
            ],
            [Path("template/init/gitattributes"), Path(".gitattributes")],
            [Path("template/init/LICENSE-CC-BY-4.0.txt"), Path("LICENSE.txt")],
            [
                Path("template/init/colrev_update.yml"),
                Path(".github/workflows/colrev_update.yml"),
            ],
        ]
        for retrieval_path, target_path in files_to_retrieve:
            colrev.env.utils.retrieve_package_file(
                template_file=retrieval_path, target=target_path
            )

        self.review_manager = colrev.review_manager.ReviewManager()

        settings = self.review_manager.settings

        committer, email = self.review_manager.get_committer()
        settings.project.authors = [
            colrev.settings.Author(
                name=committer,
                initials="".join(part[0] for part in committer.split(" ")),
                email=email,
            )
        ]

        colrev_version = version("colrev")
        if "+" in colrev_version:
            colrev_version = colrev_version[: colrev_version.rfind("+")]
        settings.project.colrev_version = colrev_version

        settings.project.title = self.title
        self.review_type = settings.project.review_type

        # Principle: adapt values provided by the default settings.json
        # instead of creating a new settings.json

        review_types = self.review_manager.get_review_types(
            review_type=self.review_type
        )
        settings = review_types.packages[self.review_type].initialize(settings=settings)

        self.review_manager.save_settings()

        project_title = self.review_manager.settings.project.title
        if "review" in project_title.lower():
            colrev.env.utils.inplace_change(
                filename=Path("readme.md"),
                old_string="{{project_title}}",
                new_string=project_title.rstrip(" ").capitalize(),
            )
        else:
            package_manager = self.review_manager.get_package_manager()
            check_operation = colrev.operation.CheckOperation(
                review_manager=self.review_manager
            )
            review_type_endpoint = package_manager.load_packages(
                package_type=colrev.env.package_manager.PackageEndpointType.review_type,
                selected_packages=[{"endpoint": self.review_type}],
                operation=check_operation,
                ignore_not_available=False,
            )
            r_type_suffix = str(review_type_endpoint[self.review_type])

            colrev.env.utils.inplace_change(
                filename=Path("readme.md"),
                old_string="{{project_title}}",
                new_string=project_title.rstrip(" ").capitalize()
                + f": A {r_type_suffix} protocol",
            )

        if self.light or platform.system() == "Darwin":
            settings.data.data_package_endpoints = [
                x
                for x in settings.data.data_package_endpoints
                if x["endpoint"] not in ["colrev.paper_md"]
            ]
            settings.sources = [
                x for x in settings.sources if x.endpoint not in ["colrev.pdfs_dir"]
            ]

            settings.pdf_prep.pdf_prep_package_endpoints = [
                x
                for x in settings.pdf_prep.pdf_prep_package_endpoints
                if x["endpoint"]
                not in [
                    "colrev.pdf_check_ocr",
                    "colrev.remove_coverpage",
                    "colrev.remove_last_page",
                    "colrev.create_tei",
                ]
            ]

        self.review_manager.save_settings()

        # Note : to avoid file setup at colrev status (calls data_operation.main)
        data_operation = self.review_manager.get_data_operation(
            notify_state_transition_operation=False
        )
        data_operation.main(silent_mode=True)

        self.review_manager.logger.info("Set up %s", self.review_type)

        for source in settings.sources:
            self.review_manager.logger.info(
                " add search %s", source.endpoint.replace("colrev.", "")
            )

        for data_package_endpoint in settings.data.data_package_endpoints:
            self.review_manager.logger.info(
                " add data   %s",
                data_package_endpoint["endpoint"].replace("colrev.", ""),
            )

        with open("data/records.bib", mode="w", encoding="utf-8") as file:
            file.write("\n")

        git_repo = self.review_manager.dataset.get_repo()
        git_repo.git.add(all=True)

    def __post_commit_edits(self) -> None:
        if self.review_type == "colrev.curated_masterdata":
            self.review_manager.logger.info("Post-commit edits")
            self.review_manager.settings.project.curation_url = "TODO"
            self.review_manager.settings.project.curated_fields = ["url", "doi", "TODO"]

            pdf_source_l = [
                s
                for s in self.review_manager.settings.sources
                if "data/search/pdfs.bib" == str(s.filename)
            ]
            if pdf_source_l:
                pdf_source = pdf_source_l[0]
                pdf_source.search_parameters = {
                    "scope": {
                        "path": "pdfs",
                        "journal": "TODO",
                        "subdir_pattern": "TODO:volume_number|year",
                    }
                }

            crossref_source_l = [
                s
                for s in self.review_manager.settings.sources
                if "data/search/CROSSREF.bib" == str(s.filename)
            ]
            if crossref_source_l:
                crossref_source = crossref_source_l[0]
                crossref_source.search_parameters = {"scope": {"journal_issn": "TODO"}}

            self.review_manager.save_settings()

            self.review_manager.logger.info("Completed setup.")

    def __setup_git(self) -> None:
        self.logger.info("Create CoLRev repository")

        self.logger.info("Set up git repository")

        git.Repo.init()

        # To check if git actors are set
        environment_manager = colrev.env.environment_manager.EnvironmentManager()
        environment_manager.get_name_mail_from_git()

        logging.info("Install latest pre-commmit hooks")
        scripts_to_call = [
            {
                "description": "Install pre-commit hooks",
                "command": ["pre-commit", "install"],
            },
            {
                "description": "",
                "command": [
                    "pre-commit",
                    "install",
                    "--hook-type",
                    "prepare-commit-msg",
                ],
            },
            {
                "description": "",
                "command": ["pre-commit", "install", "--hook-type", "pre-push"],
            },
            {"description": "", "command": ["pre-commit", "autoupdate"]},
            {"description": "", "command": ["daff", "git", "csv"]},
        ]
        for script_to_call in scripts_to_call:
            try:
                if script_to_call["description"]:
                    self.logger.debug("%s...", script_to_call["description"])
                check_call(
                    script_to_call["command"], stdout=DEVNULL, stderr=STDOUT
                )  # nosec
            except CalledProcessError:
                if " ".join(script_to_call["command"]) == "pre-commit autoupdate":
                    pass
                else:
                    self.logger.error(
                        "%sFailed: %s%s",
                        colors.RED,
                        " ".join(script_to_call),
                        colors.END,
                    )

    def __create_example_repo(self) -> None:
        """The example repository is intended to provide an initial illustration
        of CoLRev. It focuses on a quick overview of the process and does
        not cover advanced features or special cases."""

        self.logger.info("Include 30_example_records.bib")
        colrev.env.utils.retrieve_package_file(
            template_file=Path("template/example/30_example_records.bib"),
            target=Path("data/search/30_example_records.bib"),
        )

        git_repo = git.Repo.init()
        git_repo.index.add(["data/search/30_example_records.bib"])

        with open("settings.json", encoding="utf-8") as file:
            settings = json.load(file)

        settings["dedupe"]["dedupe_package_endpoints"] = [
            {"endpoint": "colrev.simple_dedupe"}
        ]

        with open("settings.json", "w", encoding="utf-8") as outfile:
            json.dump(settings, outfile, indent=4)
        git_repo.index.add(["settings.json"])

    def __create_local_pdf_collection(self) -> None:
        self.review_manager.report_logger.handlers = []

        local_index = self.review_manager.get_local_index()
        local_pdf_collection_path = local_index.local_environment_path / Path(
            "local_pdf_collection"
        )

        if not local_pdf_collection_path.is_dir():
            local_pdf_collection_path.mkdir(parents=True, exist_ok=True)
            os.chdir(local_pdf_collection_path)
            Initializer(
                review_type="colrev.literature_review",
                local_pdf_collection=True,
            )
            self.logger.info("Created local_pdf_collection repository")

        os.chdir(self.target_path)



#! /usr/bin/env python
"""Quality model for records."""
from __future__ import annotations

import importlib
from pathlib import Path

import colrev.qm.checkers
import colrev.record


class QualityModel:
    """The quality model for records"""

    __checker_path = Path(__file__).parent / Path("checkers")
    checkers = []  # type: ignore

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        self.review_manager = review_manager
        self.defects_to_ignore = self.review_manager.settings.prep.defects_to_ignore
        self.__register_checkers()

    def __register_checkers(self) -> None:
        """Register checkers from the checker directory, looking for a
        'register' function in each one.
        """

        for filename in self.__checker_path.glob("*.py"):
            if "__init__" in str(filename):
                continue

            try:
                module = importlib.import_module("colrev.qm.checkers." + filename.stem)
            except ValueError as exc:
                print(exc)
            except ImportError as exc:
                print(
                    f"Problem importing module {filename}: {exc}"
                )  # , file=sys.stderr
            else:
                if hasattr(module, "register"):
                    module.register(self)

    def register_checker(self, checker) -> None:  # type: ignore
        """Register a checker"""
        self.checkers.append(checker)

    def run(self, *, record: colrev.record.Record) -> None:
        """Run the checkers"""
        for checker in self.checkers:
            if checker.msg in self.defects_to_ignore:
                continue
            checker.run(record=record)



#! /usr/bin/env python3
"""CoLRev status operation: Display the project status."""
from __future__ import annotations

import io
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import yaml

import colrev.env.utils
import colrev.operation
import colrev.record
import colrev.ui_cli.cli_colors as colrev_colors


class Status(colrev.operation.Operation):
    """Determine the status of the project"""

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=colrev.operation.OperationsType.check,
        )

    def get_analytics(self) -> dict:
        """Get status analytics"""

        analytics_dict = {}
        git_repo = self.review_manager.dataset.get_repo()

        revlist = list(
            (
                commit.hexsha,
                commit.author.name,
                commit.committed_date,
                (commit.tree / "status.yaml").data_stream.read(),
            )
            for commit in git_repo.iter_commits(paths="status.yaml")
        )
        for ind, (commit_id, commit_author, committed_date, filecontents) in enumerate(
            revlist
        ):
            try:
                var_t = io.StringIO(filecontents.decode("utf-8"))

                # TBD: we could simply include the whole status.yaml
                # (to create a general-purpose status analyzer)
                # -> flatten nested structures (e.g., overall/currently)
                # -> integrate with get_status (current data) -
                # and get_prior? (levels: aggregated_statistics vs. record-level?)

                data_loaded = yaml.safe_load(var_t)
                analytics_dict[len(revlist) - ind] = {
                    "atomic_steps": data_loaded["atomic_steps"],
                    "completed_atomic_steps": data_loaded["completed_atomic_steps"],
                    "commit_id": commit_id,
                    "commit_author": commit_author,
                    "committed_date": committed_date,
                    "search": data_loaded["overall"]["md_retrieved"],
                    "included": data_loaded["overall"]["rev_included"],
                }
            except (IndexError, KeyError):
                pass

        # keys = list(analytics_dict.values())[0].keys()
        # with open("analytics.csv", "w", newline="", encoding="utf8") as output_file:
        #     dict_writer = csv.DictWriter(output_file, keys)
        #     dict_writer.writeheader()
        #     dict_writer.writerows(reversed(analytics_dict.values()))

        return analytics_dict

    def get_review_status_report(
        self, *, records: Optional[dict] = None, colors: bool = True
    ) -> str:
        """Get the review status report"""

        status_stats = self.review_manager.get_status_stats(records=records)

        template = colrev.env.utils.get_template(
            template_path="template/ops/status.txt"
        )

        if colors:
            content = template.render(status_stats=status_stats, colors=colrev_colors)
        else:
            content = template.render(status_stats=status_stats, colors=None)

        return content


@dataclass
class StatusStats:
    """Data class for status statistics"""

    # pylint: disable=too-many-instance-attributes
    atomic_steps: int
    nr_curated_records: int
    currently: StatusStatsCurrently
    overall: StatusStatsOverall
    completed_atomic_steps: int
    completeness_condition: bool

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        records: Optional[dict] = None,
    ) -> None:
        self.review_manager = review_manager
        colrev.operation.CheckOperation(review_manager=review_manager)

        if records:
            self.records = records
        else:
            self.records = self.review_manager.dataset.load_records_dict()

        self.status_list = [x["colrev_status"] for x in self.records.values()]
        self.screening_criteria = [
            x["screening_criteria"]
            for x in self.records.values()
            if x.get("screening_criteria", "") not in ["", "NA"]
        ]

        self.md_duplicates_removed = 0
        for item in self.records.values():
            self.md_duplicates_removed += (
                len([o for o in item["colrev_origin"] if not o.startswith("md_")]) - 1
            )

        origin_list = [x["colrev_origin"] for x in self.records.values()]
        self.record_links = 0
        for origin in origin_list:
            self.record_links += len([o for o in origin if not o.startswith("md_")])

        criteria = list(review_manager.settings.screen.criteria.keys())
        self.screening_statistics = {crit: 0 for crit in criteria}
        for screening_case in self.screening_criteria:
            for criterion in screening_case.split(";"):
                criterion_name, decision = criterion.split("=")
                if decision == "out":
                    self.screening_statistics[criterion_name] += 1

        self.currently = self.StatusStatsCurrently(status_stats=self)
        self.overall = self.StatusStatsOverall(status_stats=self)

        self.completed_atomic_steps = 0
        self.nr_incomplete = 0

        self.__overall_stats_backward_calculation()

        self.currently.non_processed = (
            self.currently.md_imported
            + self.currently.md_retrieved
            + self.currently.md_needs_manual_preparation
            + self.currently.md_prepared
        )

        self.currently.md_retrieved = self.overall.md_retrieved - self.record_links

        self.completeness_condition = (
            (0 == self.nr_incomplete)
            and (0 == self.currently.md_retrieved)
            and self.overall.md_retrieved > 0
        )

        self.currently.exclusion = self.screening_statistics

        self.overall.rev_screen = self.overall.pdf_prepared

        self.overall.rev_prescreen = self.overall.md_processed
        self.currently.pdf_needs_retrieval = self.currently.rev_prescreen_included

        colrev_masterdata_items = [
            x["colrev_masterdata_provenance"]
            for x in self.records.values()
            if "colrev_masterdata_provenance" in x
        ]

        self.nr_curated_records = len(
            [x for x in colrev_masterdata_items if "CURATED" in x]
        )
        if review_manager.settings.is_curated_masterdata_repo():
            self.nr_curated_records = self.overall.md_processed

        self.atomic_steps = (
            # initially, all records have to pass 8 operations
            8 * self.overall.md_retrieved
            # for removed duplicates, 5 operations are no longer needed
            - 5 * self.currently.md_duplicates_removed
            # for rev_prescreen_excluded, 4 operations are no longer needed
            - 4 * self.currently.rev_prescreen_excluded
            - 3 * self.currently.pdf_not_available
            - self.currently.rev_excluded
        )

        self.perc_curated = 0
        denominator = (
            self.overall.md_processed
            + self.currently.md_prepared
            + self.currently.md_needs_manual_preparation
            + self.currently.md_imported
        )

        if denominator > 0:
            self.perc_curated = int((self.nr_curated_records / (denominator)) * 100)

    def __overall_stats_backward_calculation(self) -> None:
        """Calculate the state_x overall stats (based on backward calculation)"""
        # self.review_manager.logger.debug(
        #     "Set overall colrev_status statistics (going backwards)"
        # )
        visited_states = []
        current_state = colrev.record.RecordState.rev_synthesized  # start with the last
        atomic_step_number = 0
        while True:
            # self.review_manager.logger.debug(
            #     "current_state: %s with %s",
            #     current_state,
            #     getattr(self.overall, str(current_state)),
            # )
            if colrev.record.RecordState.md_prepared == current_state:
                overall_md_prepared = (
                    getattr(self.overall, str(current_state))
                    + self.md_duplicates_removed
                )
                getattr(self.overall, str(current_state), overall_md_prepared)

            states_to_consider = [current_state]
            predecessors: list[dict[str, typing.Any]] = [
                {
                    "trigger": "init",
                    "source": colrev.record.RecordState.md_imported,
                    "dest": colrev.record.RecordState.md_imported,
                }
            ]
            # Go backward through the process model
            predecessor = None
            while predecessors:
                predecessors = [
                    t
                    for t in colrev.record.RecordStateModel.transitions
                    if t["source"] in states_to_consider
                    and t["dest"] not in visited_states
                ]
                for predecessor in predecessors:
                    # self.review_manager.logger.debug(
                    #     " add %s from %s (predecessor transition: %s)",
                    #     getattr(self.overall, str(predecessor["dest"])),
                    #     str(predecessor["dest"]),
                    #     predecessor["trigger"],
                    # )
                    setattr(
                        self.overall,
                        str(current_state),
                        (
                            getattr(self.overall, str(current_state))
                            + getattr(self.overall, str(predecessor["dest"]))
                        ),
                    )
                    visited_states.append(predecessor["dest"])
                    if predecessor["dest"] not in states_to_consider:
                        states_to_consider.append(predecessor["dest"])
                if len(predecessors) > 0:
                    if predecessors[0]["trigger"] != "init":
                        # ignore _man versions to avoid double-counting:
                        if "_man" not in predecessors[0]["trigger"]:
                            self.completed_atomic_steps += getattr(
                                self.overall, str(predecessor["dest"])
                            )
                        # Note : load is not a predecessor so we need to
                        # correct for a missing step (same number like prep)
                        if predecessors[0]["trigger"] == "prep":
                            self.completed_atomic_steps += getattr(
                                self.overall, str(predecessor["dest"])
                            )

            atomic_step_number += 1
            # Note : the following does not consider multiple parallel steps.
            for trans_for_completeness in [
                t
                for t in colrev.record.RecordStateModel.transitions
                if current_state == t["dest"]
            ]:
                self.nr_incomplete += getattr(
                    self.currently, str(trans_for_completeness["source"])
                )

            t_list = [
                t
                for t in colrev.record.RecordStateModel.transitions
                if current_state == t["dest"]
            ]
            transition: dict = t_list.pop()
            if current_state == colrev.record.RecordState.md_imported:
                break
            current_state = transition["source"]  # go a step back
            self.currently.non_completed += getattr(self.currently, str(current_state))

    def get_active_metadata_operation_info(self) -> str:
        """Get active metadata operation info (convenience function for status printing)"""
        infos = []
        if self.currently.md_retrieved > 0:
            infos.append(f"{self.currently.md_retrieved} to load")
        if self.currently.md_imported > 0:
            infos.append(f"{self.currently.md_imported} to prepare")
        if self.currently.md_needs_manual_preparation > 0:
            infos.append(
                f"{self.currently.md_needs_manual_preparation} to prepare manually"
            )
        if self.currently.md_prepared > 0:
            infos.append(f"{self.currently.md_prepared} to deduplicate")
        return ", ".join(infos)

    def get_active_pdf_operation_info(self) -> str:
        """Get active PDF operation info (convenience function for status printing)"""
        infos = []
        if self.currently.rev_prescreen_included > 0:
            infos.append(f"{self.currently.rev_prescreen_included} to retrieve")
        if self.currently.pdf_needs_manual_retrieval > 0:
            infos.append(
                f"{self.currently.pdf_needs_manual_retrieval} to retrieve manually"
            )
        if self.currently.pdf_imported > 0:
            infos.append(f"{self.currently.pdf_imported} to prepare")
        if self.currently.pdf_needs_manual_preparation > 0:
            infos.append(
                f"{self.currently.pdf_needs_manual_preparation} to prepare manually"
            )
        return ", ".join(infos)

    def get_transitioned_records(
        self, current_origin_states_dict: dict
    ) -> list[typing.Dict]:
        """Get the transitioned records"""

        committed_origin_states_dict = (
            self.review_manager.dataset.get_committed_origin_state_dict()
        )
        transitioned_records = []
        for (
            committed_origin,
            committed_colrev_status,
        ) in committed_origin_states_dict.items():
            transitioned_record = {
                "origin": committed_origin,
                "source": committed_colrev_status,
                "dest": current_origin_states_dict.get(
                    committed_origin, "no_source_state"
                ),
            }

            operations_type = [
                x["trigger"]
                for x in colrev.record.RecordStateModel.transitions
                if x["source"] == transitioned_record["source"]
                and x["dest"] == transitioned_record["dest"]
            ]
            if (
                len(operations_type) == 0
                and transitioned_record["source"] != transitioned_record["dest"]
            ):
                transitioned_record["operations_type"] = "invalid_transition"

            if len(operations_type) > 0:
                transitioned_record["operations_type"] = operations_type[0]
                transitioned_records.append(transitioned_record)

        return transitioned_records

    def get_priority_operations(self, *, current_origin_states_dict: dict) -> list:
        """Get the priority operations"""

        # get "earliest" states (going backward)
        earliest_state = []
        search_states = [colrev.record.RecordState.rev_synthesized]
        while True:
            if any(
                search_state in current_origin_states_dict.values()
                for search_state in search_states
            ):
                earliest_state = [
                    search_state
                    for search_state in search_states
                    if search_state in current_origin_states_dict.values()
                ]
            search_states = [
                x["source"]  # type: ignore
                for x in colrev.record.RecordStateModel.transitions
                if x["dest"] in search_states
            ]
            if [] == search_states:
                break
        # print(f'earliest_state: {earliest_state}')

        # next: get the priority transition for the earliest states
        priority_transitions = [
            x["trigger"]
            for x in colrev.record.RecordStateModel.transitions
            if x["source"] in earliest_state
        ]

        priority_operations = list(set(priority_transitions))

        self.review_manager.logger.debug(f"priority_operations: {priority_operations}")
        return priority_operations

    def get_active_operations(self, *, current_origin_states_dict: dict) -> list:
        """Get the active processing functions"""

        active_operations: typing.List[str] = []
        for state in set(current_origin_states_dict.values()):
            valid_transitions = colrev.record.RecordStateModel.get_valid_transitions(
                state=state
            )
            active_operations.extend(valid_transitions)

        self.review_manager.logger.debug(f"active_operations: {set(active_operations)}")
        return active_operations

    def get_operation_in_progress(self, *, transitioned_records: list) -> list:
        """Get the operation currently in progress"""

        in_progress_operation = list(
            {x["operations_type"] for x in transitioned_records}
        )
        self.review_manager.logger.debug(
            f"in_progress_operation: {in_progress_operation}"
        )
        return in_progress_operation

    @dataclass
    class StatusStatsParent:
        """Parent class for StatusStatsCurrently and StatusStatsOverall"""

        # pylint: disable=too-many-instance-attributes
        # Note : StatusStatsCurrently and StatusStatsOverall start with the same frequencies
        def __init__(
            self,
            *,
            status_stats: StatusStats,
        ) -> None:
            self.status_stats = status_stats

            self.md_retrieved = self.__get_freq(colrev.record.RecordState.md_retrieved)

            self.md_imported = self.__get_freq(colrev.record.RecordState.md_imported)
            self.md_needs_manual_preparation = self.__get_freq(
                colrev.record.RecordState.md_needs_manual_preparation
            )
            self.md_prepared = self.__get_freq(colrev.record.RecordState.md_prepared)
            self.md_processed = self.__get_freq(colrev.record.RecordState.md_processed)
            self.rev_prescreen_excluded = self.__get_freq(
                colrev.record.RecordState.rev_prescreen_excluded
            )
            self.rev_prescreen_included = self.__get_freq(
                colrev.record.RecordState.rev_prescreen_included
            )
            self.pdf_needs_manual_retrieval = self.__get_freq(
                colrev.record.RecordState.pdf_needs_manual_retrieval
            )
            self.pdf_imported = self.__get_freq(colrev.record.RecordState.pdf_imported)
            self.pdf_not_available = self.__get_freq(
                colrev.record.RecordState.pdf_not_available
            )
            self.pdf_needs_manual_preparation = self.__get_freq(
                colrev.record.RecordState.pdf_needs_manual_preparation
            )
            self.pdf_prepared = self.__get_freq(colrev.record.RecordState.pdf_prepared)
            self.rev_excluded = self.__get_freq(colrev.record.RecordState.rev_excluded)
            self.rev_included = self.__get_freq(colrev.record.RecordState.rev_included)
            self.rev_synthesized = self.__get_freq(
                colrev.record.RecordState.rev_synthesized
            )
            self.md_duplicates_removed = self.status_stats.md_duplicates_removed

        def __get_freq(self, colrev_status: colrev.record.RecordState) -> int:
            return len([x for x in self.status_stats.status_list if colrev_status == x])

    @dataclass
    class StatusStatsCurrently(StatusStatsParent):
        """The current status statistics"""

        # pylint: disable=too-many-instance-attributes
        md_retrieved: int
        md_imported: int
        md_prepared: int
        md_needs_manual_preparation: int
        md_duplicates_removed: int
        md_processed: int
        non_processed: int
        rev_prescreen_excluded: int
        rev_prescreen_included: int
        pdf_needs_retrieval: int
        pdf_needs_manual_retrieval: int
        pdf_not_available: int
        pdf_imported: int
        pdf_needs_manual_preparation: int
        pdf_prepared: int
        rev_excluded: int
        rev_included: int
        rev_synthesized: int
        non_completed: int
        exclusion: dict

        def __init__(
            self,
            *,
            status_stats: StatusStats,
        ) -> None:
            self.exclusion: typing.Dict[str, int] = {}
            self.non_completed = 0
            self.non_processed = 0
            super().__init__(status_stats=status_stats)
            self.pdf_needs_retrieval = self.rev_prescreen_included

    @dataclass
    class StatusStatsOverall(StatusStatsParent):
        """The overall-status statistics (records currently/previously in each state)"""

        # pylint: disable=too-many-instance-attributes
        md_retrieved: int
        md_imported: int
        md_needs_manual_preparation: int
        md_prepared: int
        md_processed: int
        rev_prescreen: int
        rev_prescreen_excluded: int
        rev_prescreen_included: int
        pdf_needs_manual_retrieval: int
        pdf_imported: int
        pdf_not_available: int
        pdf_needs_manual_preparation: int
        pdf_prepared: int
        rev_excluded: int
        rev_included: int
        rev_screen: int
        rev_synthesized: int

        def __init__(
            self,
            *,
            status_stats: StatusStats,
        ) -> None:
            self.rev_screen = 0
            self.rev_prescreen = 0
            super().__init__(status_stats=status_stats)
            self.md_retrieved = self.__get_nr_search(
                search_dir=self.status_stats.review_manager.search_dir
            )

        def __get_nr_search(self, *, search_dir: Path) -> int:
            if not search_dir.is_dir():
                return 0
            bib_files = search_dir.glob("*.bib")
            number_search = 0
            for search_file in bib_files:
                # Note : skip md-prep sources
                if str(search_file.name).startswith("md_"):
                    continue
                number_search += self.status_stats.review_manager.dataset.get_nr_in_bib(
                    file_path=search_file
                )
            return number_search
